# Copyright the author(s) of DLK.
# Copyright spark-nlp-display.
#
# This original source code is copied from spark-nlp-display and licensed under the Apache license.

import json
import math
import os
import random
import re
from typing import List

import numpy as np
import svgwrite

from dlk.utils.display.annotation import SpanAnnotation

# overlap_hist = []
# y_hist_dict = {}
x_i_diff_dict = {}
x_o_diff_dict = {}


class RelationExtractionVisualizer:
    def __init__(self, exclude_relations: List = [], width=900):
        here = os.path.abspath(os.path.dirname(__file__))
        self.width = width
        self.exclude_relations = set(exclude_relations)
        with open(
            os.path.join(here, "label_colors/relations.json"), "r", encoding="utf-8"
        ) as f_:
            self.color_dict = json.load(f_)
        with open(
            os.path.join(here, "label_colors/ner.json"), "r", encoding="utf-8"
        ) as f_:
            self.entity_color_dict = json.load(f_)
            self.entity_color_dict = dict(
                (k.lower(), v) for k, v in self.entity_color_dict.items()
            )
            self.font_path = os.path.join(here, "fonts/Lucida_Console.ttf")
            self.main_font = "Lucida"

    def __get_color(self, l):
        r = lambda: random.randint(0, 200)
        return "#%02X%02X%02X" % (r(), r(), r())

    def __size(self, text):
        # NOTE: Chinese characters length is 16(experimental)
        chinese_cnt = len(re.findall(r"[\u4e00-\u9fff]", text))
        return ((len(text) + 1 - chinese_cnt) * 9.7) + (chinese_cnt * 16) - 5
        # return ((len(text) + 1) * 9.7) - 5

    def __draw_line(
        self,
        dwg,
        s_x,
        s_y,
        e_x,
        e_y,
        d_type,
        color,
        show_relations,
        size_of_entity_label,
    ):
        eps = 0.0000001

        def get_bezier_coef(points):
            # since the formulas work given that we have n+1 points
            # then n must be this:
            n = len(points) - 1

            # build coefficents matrix
            C = 4 * np.identity(n)
            np.fill_diagonal(C[1:], 1)
            np.fill_diagonal(C[:, 1:], 1)
            C[0, 0] = 2
            C[n - 1, n - 1] = 7
            C[n - 1, n - 2] = 2

            # build points vector
            P = [2 * (2 * points[i] + points[i + 1]) for i in range(n)]
            P[0] = points[0] + 2 * points[1]
            P[n - 1] = 8 * points[n - 1] + points[n]

            # solve system, find a & b
            A = np.linalg.solve(C, P)
            B = [0] * n
            for i in range(n - 1):
                B[i] = 2 * points[i + 1] - A[i + 1]
            B[n - 1] = (A[n - 1] + points[n]) / 2

            return A, B

        # returns the general Bezier cubic formula given 4 control points
        def get_cubic(a, b, c, d):
            return (
                lambda t: np.power(1 - t, 3) * a
                + 3 * np.power(1 - t, 2) * t * b
                + 3 * (1 - t) * np.power(t, 2) * c
                + np.power(t, 3) * d
            )

        # return one cubic curve for each consecutive points
        def get_bezier_cubic(points):
            A, B = get_bezier_coef(points)
            return [
                get_cubic(points[i], A[i], B[i], points[i + 1])
                for i in range(len(points) - 1)
            ]

        # evalute each cubic curve on the range [0, 1] sliced in n points
        def evaluate_bezier(points, n):
            curves = get_bezier_cubic(points)
            return np.array([fun(t) for fun in curves for t in np.linspace(0, 1, n)])

        def draw_pointer(dwg, s_x, s_y, e_x, e_y):
            size = 5
            ratio = 1
            fullness1 = 2
            fullness2 = 3
            bx = e_x
            ax = s_x
            by = e_y
            ay = s_y
            abx = bx - ax
            aby = by - ay
            ab = np.sqrt(abx * abx + aby * aby) + eps

            cx = bx - size * abx / ab
            cy = by - size * aby / ab
            dx = cx + (by - cy) / ratio
            dy = cy + (cx - bx) / ratio
            ex = cx - (by - cy) / ratio
            ey = cy - (cx - bx) / ratio
            fx = (fullness1 * cx + bx) / fullness2
            fy = (fullness1 * cy + by) / fullness2

            text_place_y = s_y - (abs(s_y - e_y) / 2)
            """
            line = dwg.add(dwg.polyline(
                      [
                      (bx, by),
                      (dx, dy),
                      (fx, fy),
                      (ex, ey),
                      (bx, by)
                      ],
                      stroke=color, stroke_width = "1", fill='none',))
            """
            line = dwg.add(
                dwg.polyline(
                    [(dx, dy), (bx, by), (ex, ey), (bx, by)],
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            return text_place_y

        unique_o_index = str(s_x) + str(s_y)
        unique_i_index = str(e_x) + str(e_y)
        if s_x > e_x:
            if unique_o_index in x_o_diff_dict:
                s_x -= 5
            else:
                s_x -= 10
                x_o_diff_dict[unique_o_index] = 5
            if s_y > e_y:
                e_x += size_of_entity_label
            elif s_y < e_y:
                s_x -= size_of_entity_label

            if unique_i_index in x_i_diff_dict:
                e_x += 5
            else:
                e_x += 10
                x_i_diff_dict[unique_i_index] = 5
        else:
            if unique_o_index in x_o_diff_dict:
                s_x += 5
            else:
                s_x += 10
                x_o_diff_dict[unique_o_index] = 5
            if s_y > e_y:
                e_x -= size_of_entity_label
            elif s_y < e_y:
                s_x += size_of_entity_label

            if unique_i_index in x_i_diff_dict:
                e_x -= 5
            else:
                e_x -= 10
                x_i_diff_dict[unique_i_index] = 5
        # this_y_vals = list(range(min(s_x,e_x), max(s_x,e_x)+1))
        # this_y_vals = [ str(s_y)+'|'+str(i) for i in this_y_vals]
        # common = set(this_y_vals) & set(overlap_hist)
        # overlap_hist.extend(this_y_vals)
        # if s_y not in y_hist_dict:
        #    y_hist_dict[s_y] = 20
        # if common:
        #    y_hist_dict[s_y] += 20
        # y_increase = y_hist_dict[s_y]
        angle = -1
        if s_y == e_y:
            angle = 0
            s_y -= 20
            e_y = s_y - 4  # 55

            text_place_y = s_y - 35

            pth = evaluate_bezier(
                np.array([[s_x, s_y], [(s_x + e_x) / 2.0, s_y - 40], [e_x, e_y]]), 50
            )
            dwg.add(
                dwg.polyline(
                    pth,
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            draw_pointer(dwg, (s_x + e_x) / 2.0, s_y - 50, e_x, e_y)
        elif s_y >= e_y:
            e_y += 15
            s_y -= 20
            text_place_y = s_y - (abs(s_y - e_y) / 2)

            pth = evaluate_bezier(
                np.array(
                    [
                        [s_x, s_y],
                        # [((3*s_x)+e_x)/4.0, (s_y+e_y)/2.0],
                        [(s_x + e_x) / 2.0, (s_y + e_y) / 2.0],
                        # [(s_x+(3*e_x))/4.0,(s_y+e_y)/2.0],
                        [e_x, e_y],
                    ]
                ),
                50,
            )
            dwg.add(
                dwg.polyline(
                    pth,
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            draw_pointer(dwg, s_x, s_y, e_x, e_y)

            """
            line = dwg.add(dwg.polyline(
                    [(s_x, s_y),(s_x, s_y-y_increase), (e_x, s_y-y_increase),
                    (e_x, e_y),
                    (e_x+2, e_y),
                    (e_x, e_y-4),
                    (e_x-2, e_y),
                    (e_x, e_y)
                    ],
                    stroke=color, stroke_width = "2", fill='none',))
            """
        else:
            s_y += 15
            e_y -= 20
            text_place_y = s_y + (abs(s_y - e_y) / 2)

            pth = evaluate_bezier(
                np.array(
                    [
                        [s_x, s_y],
                        # [((3*s_x)+e_x)/4.0, (s_y+e_y)/2.0],
                        [(s_x + e_x) / 2.0, (s_y + e_y) / 2.0],
                        # [(s_x+(3*e_x))/4.0,(s_y+e_y)/2.0],
                        [e_x, e_y],
                    ]
                ),
                50,
            )
            dwg.add(
                dwg.polyline(
                    pth,
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            draw_pointer(dwg, s_x, s_y, e_x, e_y)

        if show_relations:
            if angle == -1:
                angle = math.degrees(math.atan((s_y - e_y) / ((s_x - e_x) + eps)))
            rel_temp_size = self.__size(d_type) / 1.35
            rect_x, rect_y = (
                ((s_x + e_x) / 2.0) - (rel_temp_size / 2.0) - 3,
                text_place_y - 10,
            )
            rect_w, rect_h = (rel_temp_size + 3, 13)
            dwg.add(
                dwg.rect(
                    insert=(rect_x, rect_y),
                    rx=2,
                    ry=2,
                    size=(rect_w, rect_h),
                    fill="white",
                    stroke=color,
                    stroke_width="1",
                    transform=f"rotate({angle} {rect_x+rect_w/2} {rect_y+rect_h/2})",
                )
            )

            dwg.add(
                dwg.text(
                    d_type,
                    insert=(((s_x + e_x) / 2) - (rel_temp_size / 2.0), text_place_y),
                    fill=color,
                    font_size="12",
                    font_family="courier",
                    transform=f"rotate({angle} {rect_x+rect_w/2} {rect_y+rect_h/2})",
                )
            )

    def __gen_graph(self, rdf, selected_text, show_relations):
        rdf = [i for i in rdf if i.result.lower().strip() not in self.exclude_relations]

        done_ent1 = {}
        done_ent2 = {}
        all_done = {}

        start_y = 75
        x_limit = self.width
        y_offset = 80

        begin_index = 0
        start_x = 10
        this_line = 0

        all_entities_index = set()
        all_entities_1_index = []
        basic_dict = {}
        relation_dict = {}
        for t in rdf:
            all_entities_index.add(int(t.metadata["entity1_begin"]))
            all_entities_index.add(int(t.metadata["entity2_begin"]))
            basic_dict[int(t.metadata["entity1_begin"])] = [
                t.metadata["entity1_begin"],
                t.metadata["entity1_end"],
                t.metadata["chunk1"],
                t.metadata["entity1"],
            ]

            basic_dict[int(t.metadata["entity2_begin"])] = [
                t.metadata["entity2_begin"],
                t.metadata["entity2_end"],
                t.metadata["chunk2"],
                t.metadata["entity2"],
            ]
            if t.metadata["entity1"].lower().strip() not in self.entity_color_dict:
                self.entity_color_dict[
                    t.metadata["entity1"].lower().strip()
                ] = self.__get_color(t.metadata["entity1"].lower().strip())
            if t.metadata["entity2"].lower().strip() not in self.entity_color_dict:
                self.entity_color_dict[
                    t.metadata["entity2"].lower().strip()
                ] = self.__get_color(t.metadata["entity2"].lower().strip())

            # all_entities_1_index.append(t[4]['entity1_begin'])
        all_entities_index = np.asarray(list(all_entities_index))
        all_entities_index = all_entities_index[np.argsort(all_entities_index)]
        dwg_rects, dwg_texts = [], []
        for ent_start_ind in all_entities_index:
            e_start_now, e_end_now, e_chunk_now, e_entity_now = basic_dict[
                ent_start_ind
            ]
            prev_text = selected_text[begin_index : int(e_start_now)]
            prev_text = re.sub(r"\s*(\n)+", r"\1", prev_text.strip(), re.MULTILINE)
            begin_index = int(e_end_now) + 1
            for line_num, line in enumerate(prev_text.split("\n")):
                if line_num != 0:
                    start_y += y_offset
                    start_x = 10
                    this_line = 0
                for word_ in line.split(" "):
                    this_size = self.__size(word_)
                    if (start_x + this_size + 10) >= x_limit:
                        start_y += y_offset
                        start_x = 10
                        this_line = 0
                    dwg_texts.append(
                        [
                            word_,
                            (start_x, start_y),
                            "#546c74",
                            "16",
                            self.main_font,
                            "font-weight:100",
                        ]
                    )
                    # dwg.add(dwg.text(word_, insert=(start_x, start_y ), fill='#546c77', font_size='16',
                    #                 font_family='Monaco', style='font-weight:lighter'))
                    start_x += this_size + 10

            this_size = self.__size(e_chunk_now)
            if (start_x + this_size + 10) >= x_limit:  # or this_line >= 2:
                start_y += y_offset
                start_x = 10
                this_line = 0

            # rectange chunk 1
            dwg_rects.append(
                [
                    (start_x - 3, start_y - 18),
                    (this_size, 25),
                    self.entity_color_dict[e_entity_now.lower().strip()],
                ]
            )
            # dwg.add(dwg.rect(insert=(start_x-3, start_y-18),rx=2,ry=2, size=(this_size,25), stroke=self.entity_color_dict[e_entity_now.lower()],
            # stroke_width='1', fill=self.entity_color_dict[e_entity_now.lower()], fill_opacity='0.2'))
            # chunk1
            dwg_texts.append(
                [
                    e_chunk_now,
                    (start_x, start_y),
                    "#546c74",
                    "16",
                    self.main_font,
                    "font-weight:100",
                ]
            )
            # dwg.add(dwg.text(e_chunk_now, insert=(start_x, start_y ), fill='#546c77', font_size='16',
            #                 font_family='Monaco', style='font-weight:lighter'))
            # entity 1
            central_point_x = start_x + (this_size / 2)
            temp_size = self.__size(e_entity_now) / 2.75
            dwg_texts.append(
                [
                    e_entity_now.upper(),
                    (central_point_x - temp_size, start_y + 20),
                    "#1f77b7",
                    "12",
                    self.main_font,
                    "font-weight:lighter",
                ]
            )
            # dwg.add(dwg.text(e_entity_now.upper(),
            #                insert=(central_point_x-temp_size, start_y+20),
            #                fill='#1f77b7', font_size='12', font_family='Monaco',
            #                style='font-weight:lighter'))

            all_done[int(e_start_now)] = [central_point_x, start_y, temp_size]
            start_x += this_size + 20
            this_line += 1

        prev_text = selected_text[begin_index:]
        prev_text = re.sub(r"\s*(\n)+", r"\1", prev_text.strip(), re.MULTILINE)
        for line_num, line in enumerate(prev_text.split("\n")):
            if line_num != 0:
                start_y += y_offset
                start_x = 10
            for word_ in line.split(" "):
                this_size = self.__size(word_)
                if (start_x + this_size) >= x_limit:
                    start_y += y_offset
                    start_x = 10
                dwg_texts.append(
                    [
                        word_,
                        (start_x, start_y),
                        "#546c77",
                        "16",
                        self.main_font,
                        "font-weight:100",
                    ]
                )
                # dwg.add(dwg.text(word_, insert=(start_x, start_y ), fill='#546c77', font_size='16',
                #                 font_family='Monaco', style='font-weight:lighter'))
                start_x += this_size + 10

        dwg = svgwrite.Drawing(
            "temp.svg", profile="full", size=(x_limit, start_y + y_offset)
        )
        dwg.embed_font(self.main_font, self.font_path)

        for crect_ in dwg_rects:
            dwg.add(
                dwg.rect(
                    insert=crect_[0],
                    rx=2,
                    ry=2,
                    size=crect_[1],
                    stroke=crect_[2],
                    stroke_width="1",
                    fill=crect_[2],
                    fill_opacity="0.2",
                )
            )

        for ctext_ in dwg_texts:
            dwg.add(
                dwg.text(
                    ctext_[0],
                    insert=ctext_[1],
                    fill=ctext_[2],
                    font_size=ctext_[3],
                    font_family=ctext_[4],
                    style=ctext_[5],
                )
            )

        relation_distances = []
        relation_coordinates = []
        for row in rdf:
            if row.result.lower().strip() not in self.color_dict:
                self.color_dict[row.result.lower().strip()] = self.__get_color(
                    row.result.lower().strip()
                )

            d_key2 = all_done[int(row.metadata["entity1_begin"])]
            d_key1 = all_done[int(row.metadata["entity2_begin"])]
            this_dist = abs(d_key2[0] - d_key1[0]) + abs(d_key2[1] - d_key1[1])
            relation_distances.append(this_dist)
            relation_coordinates.append((d_key2, d_key1, row.result))

        relation_distances = np.array(relation_distances)
        relation_coordinates = np.array(relation_coordinates, dtype=object)
        temp_ind = np.argsort(relation_distances)
        relation_distances = relation_distances[temp_ind]
        relation_coordinates = relation_coordinates[temp_ind]
        for row in relation_coordinates:
            # if int(row[0][1]) == int(row[1][1]):
            size_of_entity_label = int(row[1][2])
            self.__draw_line(
                dwg,
                int(row[0][0]),
                int(row[0][1]),
                int(row[1][0]),
                int(row[1][1]),
                row[2],
                self.color_dict[row[2].lower().strip()],
                show_relations,
                size_of_entity_label,
            )

        return dwg.tostring()

    def display(self, result):
        """Displays Relation Extraction visualization.
        Inputs:
        result -- A Dataframe or dictionary.
        Output: Visualization
        """

        original_text = result["sentence"]
        entities = {}
        for entity in result["predict_entities_info"]:
            entity["chunk"] = original_text[entity["start"] : entity["end"]]
            entities[entity["entity_id"]] = entity
        relations = []
        for relation in result["predict_relations_info"]:
            relations.append(
                SpanAnnotation._from_dict(
                    {
                        "result": relation["labels"][0],
                        "metadata": {
                            "entity1": entities[relation["from"]]["labels"][0],
                            "entity1_begin": entities[relation["from"]]["start"],
                            "entity1_end": entities[relation["from"]]["end"],
                            "chunk1": entities[relation["from"]]["chunk"],
                            "entity2": entities[relation["to"]]["labels"][0],
                            "entity2_begin": entities[relation["to"]]["start"],
                            "entity2_end": entities[relation["to"]]["end"],
                            "chunk2": entities[relation["to"]]["chunk"],
                        },
                    }
                )
            )

        html_content = self.__gen_graph(relations, original_text, True)

        return html_content


if __name__ == "__main__":
    vis = RelationExtractionVisualizer()
    result = {
        "sentence": "germany 's relations with russia were never likely to be as cozy under angela merkel as under her predecessor , gerhard schr\u00f6der , who adopted a 3-year-old russian girl and , on his 60th birthday , invited president vladimir v. putin home to celebrate .",
        "uuid": "16caf15f-e6d0-11ee-98a1-5bbc69525069",
        "predict_entities_info": [
            {
                "start": 0,
                "end": 7,
                "labels": ["person"],
                "sub_token_start": 1,
                "sub_token_end": 3,
                "entity_id": "cef224e8-e79a-11ee-ae44-18c04d299e80",
            },
            {
                "start": 26,
                "end": 32,
                "labels": ["ethnicity"],
                "sub_token_start": 8,
                "sub_token_end": 10,
                "entity_id": "cef22d44-e79a-11ee-ae44-18c04d299e80",
            },
            {
                "start": 112,
                "end": 128,
                "labels": ["people"],
                "sub_token_start": 30,
                "sub_token_end": 37,
                "entity_id": "cef240ea-e79a-11ee-ae44-18c04d299e80",
            },
            {
                "start": 156,
                "end": 163,
                "labels": ["people"],
                "sub_token_start": 47,
                "sub_token_end": 49,
                "entity_id": "cef24a9a-e79a-11ee-ae44-18c04d299e80",
            },
        ],
        "predict_relations_info": [
            {
                "from": "cef24a9a-e79a-11ee-ae44-18c04d299e80",
                "to": "cef22d44-e79a-11ee-ae44-18c04d299e80",
                "labels": ["geographic_distribution"],
            },
            {
                "from": "cef240ea-e79a-11ee-ae44-18c04d299e80",
                "to": "cef224e8-e79a-11ee-ae44-18c04d299e80",
                "labels": ["nationality"],
            },
        ],
    }
    # dis = vis.display(result)
    # with open("/home/sun/Downloads/dis.html", "w") as f_:



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

logger = logging.getLogger(__name__)


@cregister("postprocessor", "span_cls")
class SpanClsPostProcessorConfig(BasePostProcessorConfig):
    """span classfication postprocessor"""

    ignore_position = BoolField(
        value=False, help="whether to ignore the position of the entity"
    )
    ignore_char = ListField(
        value=[],
        suggestions=[[" ", "(", ")", "[", "]", "-", ".", ",", ":", "'", '"']],
        help="ignore the provided char if these char is prefix or suffix of the entity",
    )
    label_vocab = StrField(value=MISSING, help="the label vocab file path")
    tokenizer_path = StrField(
        value=MISSING, help="the tokenizer file path, is not effected by `meta_dir`"
    )

    class InputMap:
        logits = StrField(value="logits", help="the output logits")
        index = StrField(value="_index", help="the index of the sample")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        sentence = StrField(value="sentence", help="the sentence of the sample")
        input_ids = StrField(value="input_ids", help="the input ids of the sample")
        entities_info = StrField(value="entities_info", help="the entities info")
        offsets = StrField(value="offsets", help="the offsets of the tokens")
        special_tokens_mask = StrField(
            value="special_tokens_mask", help="the special tokens mask"
        )
        word_ids = StrField(value="word_ids", help="the word ids")

    origin_input_map = NestField(
        value=OriginInputMap,
        help="the origin input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )
    ignore_labels = ListField(
        value=["O", "X", "S", "E"],
        help="the ignore labels, if the entity label in this list, we will ignore this entity",
    )


@register("postprocessor", "span_cls")
class SpanClsPostProcessor(BasePostProcessor):
    """PostProcess for sequence labeling task"""

    def __init__(self, config: SpanClsPostProcessorConfig):
        super(SpanClsPostProcessor, self).__init__(config)
        self.config = config
        self.label_vocab = Vocabulary.load_from_file(
            os.path.join(self.config.meta_dir, self.config.label_vocab)
        )
        with open(self.config.tokenizer_path, "r", encoding="utf-8") as f:
            tokenizer_str = json.dumps(json.load(f))
        self.tokenizer = Tokenizer.from_str(tokenizer_str)

    def _process4predict(
        self, predict_logits: torch.FloatTensor, index: int, origin_data: pd.DataFrame
    ) -> Dict:
        """gather the predict and origin text and ground_truth_entities_info for predict

        Args:
            predict: the predict span logits
            index: the data index in origin_data
            origin_data: the origin pd.DataFrame

        Returns:
            >>> one_ins info
            >>> {
            >>>     "sentence": "...",
            >>>     "uuid": "..",
            >>>     "entities_info": [".."],
            >>>     "predict_entities_info": [".."],
            >>> }

        """

        def _get_entity_info(
            sub_tokens_index: List, offset_mapping: List, word_ids: List, label: str
        ) -> Dict:
            """gather sub_tokens to get the start and end

            Args:
                sub_tokens_index: the entity tokens index list
                offset_mapping: every token offset in text
                word_ids: every token in the index of words
                label: predict label

            Returns:
                entity_info

            """
            if not sub_tokens_index or not label:
                return {}
            start = offset_mapping[sub_tokens_index[0]][0]
            end = offset_mapping[sub_tokens_index[-1]][1]
            return {"start": start, "end": end, "labels": [label]}

        one_ins = {}
        origin_ins = origin_data.iloc[int(index)]
        one_ins["sentence"] = origin_ins[self.config.origin_input_map.sentence]
        one_ins["uuid"] = origin_ins[self.config.origin_input_map.uuid]
        one_ins["entities_info"] = origin_ins[
            self.config.origin_input_map.entities_info
        ]

        word_ids = origin_ins[self.config.origin_input_map.word_ids]
        rel_token_len = len(word_ids)
        offset_mapping = origin_ins[self.config.origin_input_map.offsets][
            :rel_token_len
        ]

        predict_entities_info = []
        predict_label_ids = predict_logits.argmax(-1).cpu().numpy()
        for i in range(rel_token_len):
            for j in range(i, rel_token_len):
                if word_ids[i] is None or word_ids[j] is None:
                    continue
                predict_label_id = predict_label_ids[i][j]
                predict_label = self.label_vocab[predict_label_id]
                if (
                    predict_label == self.label_vocab.pad
                    or predict_label == self.label_vocab.unknown
                ):
                    continue
                else:
                    entity_info = _get_entity_info(
                        [i, j], offset_mapping, word_ids, predict_label
                    )
                    if entity_info:
                        predict_entities_info.append(entity_info)
        one_ins["predict_entities_info"] = predict_entities_info
        return one_ins

    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts

        """
        predicts = []
        if self.config.origin_input_map.sentence not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.sentence} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.sentence} must be provided"
            )
        if self.config.origin_input_map.uuid not in origin_data:
            logger.error(f"{self.config.origin_input_map.uuid} not in the origin data")
            raise PermissionError(
                f"{self.config.origin_input_map.uuid} must be provided"
            )

        predicts = []
        for outputs in list_batch_outputs:
            batch_logits = outputs[self.config.input_map.logits]
            # batch_special_tokens_mask = outputs[self.config.special_tokens_mask]

            indexes = list(outputs[self.config.input_map.index])

            for i, (predict, index) in enumerate(zip(batch_logits, indexes)):
                one_ins = self._process4predict(predict, index, origin_data)
                one_ins["predict_extend_return"] = self.gather_predict_extend_data(
                    outputs, i, self.config.predict_extend_return
                )
                predicts.append(one_ins)
        return predicts

    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores, recall, precision, f1

        """

        def _group_entities_info(entities_info: List[Dict], text: str) -> Dict:
            """gather the same labeled entity to the same list

            Args:
                entities_info:
                    >>> [
                    >>>     {
                    >>>         "start": start1,
                    >>>         "end": end1,
                    >>>         "labels": ["label_1"]
                    >>>     },
                    >>>     {
                    >>>         "start": start2,
                    >>>         "end": end2,
                    >>>         "labels": ["label_2"]
                    >>>     },....
                    >>> ]
                text: be labeled text

            Returns:
                >>> { "label_1" [text[start1:end1]], "label_2": [text[start_2: end_2]]...}

            """
            info = {}
            for item in entities_info:
                label = item["labels"][0]
                if label not in info:
                    info[label] = []
                start_position, end_position = item["start"], item["end"]
                while start_position < end_position:
                    if text[start_position] in self.config.ignore_char:
                        start_position += 1
                    else:
                        break
                while start_position < end_position:
                    if text[end_position - 1] in self.config.ignore_char:
                        end_position -= 1
                    else:
                        break
                if (
                    start_position == end_position
                ):  # if the entity after remove ignore char be null, we set it to origin
                    start_position, end_position = item["start"], item["end"]

                if self.config.ignore_position:
                    info[label].append(text[item["start"] : item["end"]].strip())
                else:
                    info[label].append((start_position, end_position))
            return info

        def _calc_score(predict_list: List, ground_truth_list: List):
            """use predict_list and ground_truth_list to calc scores

            Args:
                predict_list: list of predict
                ground_truth_list: list of ground_truth

            Returns:
                precision, recall, f1

            """
            category_tp = {}
            category_fp = {}
            category_fn = {}

            def _care_div(a, b):
                """return a/b or 0.0 if b == 0"""
                if b == 0:
                    return 0.0
                return a / b

            def _calc_num(_pred: List, _ground_truth: List):
                """calc tp, fn, fp

                Args:
                    pred: pred list
                    ground_truth: groud truth list

                Returns:
                    tp, fn, fp

                """
                num_p = len(_pred)
                num_t = len(_ground_truth)
                truth = 0
                for p in _pred:
                    if p in _ground_truth:
                        truth += 1
                return truth, num_t - truth, num_p - truth

            for predict, ground_truth in zip(predict_list, ground_truth_list):
                keys = set(list(predict.keys()) + list(ground_truth.keys()))
                for key in keys:
                    tp, fn, fp = _calc_num(
                        predict.get(key, []), ground_truth.get(key, [])
                    )
                    category_tp[key] = category_tp.get(key, 0) + tp
                    category_fn[key] = category_fn.get(key, 0) + fn
                    category_fp[key] = category_fp.get(key, 0) + fp

            all_tp, all_fn, all_fp = 0, 0, 0
            for key in category_tp:
                tp, fn, fp = category_tp[key], category_fn[key], category_fp[key]
                all_tp += tp
                all_fn += fn
                all_fp += fp
                precision = _care_div(tp, tp + fp)
                recall = _care_div(tp, tp + fn)
                f1 = _care_div(2 * precision * recall, precision + recall)
                logger.info(
                    f"For entity 「{key}」, the precision={precision*100 :.2f}%, the recall={recall*100:.2f}%, f1={f1*100:.2f}%"
                )
            precision = _care_div(all_tp, all_tp + all_fp)
            recall = _care_div(all_tp, all_tp + all_fn)
            f1 = _care_div(2 * precision * recall, precision + recall)
            return precision, recall, f1

        all_predicts = []
        all_ground_truths = []
        for predict in predicts:
            text = predict["sentence"]
            predict_ins = _group_entities_info(predict["predict_entities_info"], text)
            ground_truth_ins = _group_entities_info(predict["entities_info"], text)
            all_predicts.append(predict_ins)
            all_ground_truths.append(ground_truth_ins)

        precision, recall, f1 = _calc_score(all_predicts, all_ground_truths)
        real_name = self.loss_name_map(stage)
        logger.info(
            f"{real_name}_precision: {precision*100}, {real_name}_recall: {recall*100}, {real_name}_f1: {f1*100}"
        )
        return {
            f"{real_name}_precision": precision * 100,
            f"{real_name}_recall": recall * 100,
            f"{real_name}_f1": f1 * 100,
        }



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

logger = logging.getLogger(__name__)


@cregister("postprocessor", "txt_cls")
class TxtClsPostProcessorConfig(BasePostProcessorConfig):
    """postprocess for text classfication"""

    label_vocab = StrField(
        value=MISSING, help="the label vocab file name, relative `meta_dir`"
    )

    class InputMap:
        logits = StrField(value="logits", help="the output logits")
        label_ids = StrField(value="label_ids", help="the label ids of the sample")
        index = StrField(value="_index", help="the index of the sample")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        sentence = StrField(
            value="sentence", help="the sentence of the sample(for data_type=single)"
        )
        sentence_a = StrField(
            value="sentence_a", help="the sentence_a of the sample(for data_type=pair)"
        )
        sentence_b = StrField(
            value="sentence_b", help="the sentence_b of the sample(for data_type=pair)"
        )

    origin_input_map = NestField(
        value=OriginInputMap,
        help="the origin input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )
    data_type = StrField(
        value="single", options=["single", "pair"], help="the data type, single or pair"
    )
    top_k = IntField(
        value=0,
        minimum=0,
        help="the top k result, if set to `0` will get all the values",
    )


@register("postprocessor", "txt_cls")
class TxtClsPostProcessor(BasePostProcessor):
    """postprocess for text classfication"""

    def __init__(self, config: TxtClsPostProcessorConfig):
        super(TxtClsPostProcessor, self).__init__(config)
        self.config = config
        self.label_vocab = Vocabulary.load_from_file(
            os.path.join(self.config.meta_dir, self.config.label_vocab)
        )
        self.top_k = config.top_k if config.top_k > 0 else self.label_vocab.word_num

    def _get_origin_data(self, one_origin: pd.Series) -> Dict:
        """

        Args:
            one_origin: the original data

        Returns:
            the gather origin data

        """
        origin = {}
        if self.config.data_type == "single":
            sentence = one_origin[self.config.origin_input_map.sentence]
            origin["sentence"] = sentence
        else:
            sentence_a = one_origin[self.config.origin_input_map.sentence_a]
            origin["sentence_a"] = sentence_a
            sentence_b = one_origin[self.config.origin_input_map.sentence_b]
            origin["sentence_b"] = sentence_b
        origin["uuid"] = one_origin[self.config.origin_input_map.uuid]
        return origin

    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts

        """
        results = []
        for outputs in list_batch_outputs:
            logits = outputs[self.config.input_map.logits].detach()
            assert len(logits.shape) == 2
            # predict_indexes = list(torch.argmax(logits, 1))
            indexes = list(outputs[self.config.input_map.index])

            if self.config.input_map.label_ids in outputs:
                label_ids = outputs[self.config.input_map.label_ids]
            else:
                label_ids = [None] * len(indexes)
            for i, (one_logits, index, label_id) in enumerate(
                zip(logits, indexes, label_ids)
            ):
                one_ins = self._get_origin_data(origin_data.iloc[int(index)])
                one_logits = torch.softmax(one_logits, -1)
                label_values, label_indeies = torch.topk(one_logits, self.top_k, dim=-1)
                predict = {}
                for i, (label_value, label_index) in enumerate(
                    zip(label_values, label_indeies)
                ):
                    label_name = self.label_vocab.get_word(int(label_index))
                    predict[i] = [label_name, float(label_value)]
                ground_truth = []
                if label_id is not None and label_id.shape:
                    for one_label_id in label_id:
                        ground_truth.append(self.label_vocab.get_word(one_label_id))
                elif label_id:
                    ground_truth.append(self.label_vocab.get_word(label_id))
                one_ins["labels"] = ground_truth
                one_ins["predicts"] = predict
                one_ins["predict_extend_return"] = self.gather_predict_extend_data(
                    outputs, i, self.config.predict_extend_return
                )
                results.append(one_ins)
        return results

    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts

            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores, acc

        """
        right_num = 0
        for one_ins in predicts:
            labels = one_ins["labels"]
            assert (
                len(labels) == 1
            ), "We currently is not support multi label in classification postprocess"
            label = labels[0]
            one_predicts = one_ins["predicts"]
            predict_label, predict_value = one_predicts[0]  # the first predict
            if label == predict_label:
                right_num += 1
        real_name = self.loss_name_map(stage)
        return {f"{real_name}_acc": right_num / len(predicts)}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Dict, List

import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from lightning import Trainer as PLTrainer
from lightning.pytorch.loggers import TensorBoardLogger
from torch.functional import Tensor

from dlk.utils.get_root import get_root
from dlk.utils.io import open
from dlk.utils.register import register, register_module_name

ogger = logging.getLogger(__name__)


@cregister("trainer", "lightning")
class LightningTrainerConfig(Base):
    """the basic trainer config. aka lightning trainer"""

    accelerator = StrField(
        value="auto",
        options=["cpu", "gpu", "tpu", "ipu", "hpu", "mps", "auto"],
        help="""
        Supports passing different accelerator types
        ("cpu", "gpu", "tpu", "ipu", "hpu", "mps", "auto") as well as custom accelerator instances. 
        """,
    )
    strategy = StrField(
        value="auto",
        options=[
            "auto",
            "ddp",
            "ddp_notebook",
            "ddp_find_unused_parameters_true",
            "ddp_spawn",
            "deepspeed",
            "deepspeed_stage_1",
            "deepspeed_stage_2",
            "deepspeed_stage_2_offload",
            "deepspeed_stage_3",
            "deepspeed_stage_3_offload",
            "deepspeed_stage_3_offload_nvme",
            "fsdp",
            "fsdp_cpu_offload",
        ],
        help="""
        Supports passing different training strategies with aliases as well custom strategies. 
        you can create a strategy by yourself and register it 
        to the strategy registry`lightning.pytorch.strategies.StrategyRegistry`.
        """,
    )
    devices = AnyField(
        value="auto",
        help="""
        The devices to use.
        Can be set to a positive number (int or str),
        a sequence of device indices (list or str),
        the value ``-1`` to indicate all available devices should be used,
        or ``"auto"`` for automatic selection based on the chosen accelerator.
        Default: ``"auto"``.
        """,
    )
    precision = StrField(
        value="32-true",
        options=[
            "64",
            "64-true",
            "32",
            "32-true",
            "16",
            "16-mixed",
            "bf16",
            "bf16-mixed",
        ],
        help=""" 
        Double precision (64, '64' or '64-true'), full precision (32, '32' or '32-true'), 16bit mixed precision (16, '16', '16-mixed') or bfloat16 mixed precision ('bf16', 'bf16-mixed').
        Can be used on CPU, GPU, TPUs, HPUs or IPUs. Default: ``'32-true'``. 
        """,
    )
    logger = BoolField(
        value=True,
        help=""" Enable logger or not. Default: ``True``. """,
    )
    fast_dev_run = BoolField(
        value=False,
        help=""" Runs 1 batch of train, val and test to find any bugs (ie: a sort of unit test). Default: ``False``. """,
    )
    max_epochs = IntField(
        value=1000,
        help="""
        Stop training once this number of epochs is reached.
        Disabled by default (None). 
        If both max_epochs and max_steps are not specified, defaults to ``max_epochs`` = 1000. 
        To enable infinite training, set ``max_epochs`` = -1. 
        """,
    )
    min_epochs = IntField(
        value=None,
        additions=[None],
        help=""" Force training for at least these many epochs. Disabled by default (None). """,
    )
    max_steps = IntField(
        value=-1,
        help=""" 
        Stop training after this number of steps.
        Disabled by default (None). 
        If both max_epochs and max_steps are not specified, defaults to ``max_epochs`` = 1000.
        To enable infinite training, set ``max_epochs`` = -1. """,
    )
    min_steps = IntField(
        value=None,
        additions=[None],
        help=""" Force training for at least these number of steps. Disabled by default (None). """,
    )
    max_time = StrField(
        value=None,
        additions=[None],
        help=""" 
        Stop training after this amount of time has passed.
        Disabled by default (None).
        The time duration can be specified in the format DD:HH:MM:SS (days, hours, minutes seconds),
        as a :class:`datetime.timedelta`, or a :class:`numpy.timedelta64`. 
        """,
    )
    limit_train_batches = FloatField(
        value=1.0,
        minimum=0.0,
        help="""How much of training dataset to check (percent). Default: ``1.0``.""",
    )
    limit_val_batches = FloatField(
        value=1.0,
        minimum=0.0,
        help=""" How much of validation dataset to check (percent). Default: ``1.0``.""",
    )
    limit_test_batches = FloatField(
        value=1.0,
        minimum=0.0,
        help=""" How much of test dataset to check (percent). Default: ``1.0``.
                                     """,
    )
    limit_predict_batches = FloatField(
        value=1.0,
        minimum=0.0,
        help=""" How much of prediction dataset to check (percent). Default: ``1.0``. """,
    )
    overfit_batches = FloatField(
        value=0.0,
        minimum=0.0,
        help="""Overfit a fraction of training/validation data (float) or a set number of batches (int). """,
    )
    val_check_interval = FloatField(
        value=1.0,
        help=""" How often to check the validation set. Use float to check within a training epoch. Default: ``1.0``. """,
    )
    check_val_every_n_epoch = IntField(
        value=1,
        additions=[None],
        help=""" 
        Perform a validation loop every `N` training epochs. 
        If ``None``, validation will be done solely based on the number of training batches, requiring ``val_check_interval`` to be an integer value. Default: ``1``. """,
    )
    num_sanity_val_steps = IntField(
        value=2,
        minimum=-1,
        help=""" 
        Sanity check runs n validation batches before starting the training routine.
        Set it to `-1` to run all batches in all validation dataloaders.
        Default: ``2``. 
        """,
    )
    log_every_n_steps = IntField(
        value=50,
        minimum=1,
        help="""How often to log within steps. Default: ``50``. """,
    )
    enable_progress_bar = BoolField(
        value=True,
        help="""Whether to enable to progress bar by default. Default: ``True``. """,
    )
    enable_model_summary = BoolField(
        value=True,
        help=""" Whether to enable model summarization by default. Default: ``True``. """,
    )
    accumulate_grad_batches = IntField(
        value=1,
        help="""Accumulates gradients over k batches before stepping the optimizer. Default: 1.""",
    )
    gradient_clip_val = FloatField(
        value=5.0,
        additions=[None],
        help=""" 
        The value at which to clip gradients.
        Passing ``gradient_clip_val=None`` disables gradient clipping.
        If using Automatic Mixed Precision (AMP), the gradients will be unscaled before..
        """,
    )
    gradient_clip_algorithm = StrField(
        value="norm",
        options=["norm", "value"],
        help=""" 
        The gradient clipping algorithm to use.
        Pass ``gradient_clip_algorithm="value"`` to clip by value, and ``gradient_clip_algorithm="norm"`` to clip by norm.
        By default it will be set to ``"norm"``.
        """,
    )
    deterministic = BoolField(
        value=False,
        help=""" 
        If ``True``, sets whether PyTorch operations must use deterministic algorithms.
        Set to ``"warn"`` to use deterministic algorithms whenever possible, throwing warnings on operations that don't support deterministic mode (requires PyTorch 1.11+).
        """,
    )
    inference_mode = BoolField(
        value=True,
        help=""" Whether to use :func:`torch.inference_mode` or :func:`torch.no_grad` during evaluation (``validate``/``test``/``predict``).""",
    )
    use_distributed_sampler = BoolField(
        value=True,
        help=""" 
        Whether to wrap the DataLoader's sampler with :class:`torch.utils.data.DistributedSampler`.
        If not specified this is toggled automatically for strategies that require it.
        By default, it will add ``shuffle=True`` for the train sampler and ``shuffle=False`` for validation/test/predict samplers.
        If you want to disable this logic, you can pass ``False`` and add your own distributed sampler in the dataloader hooks.
        If ``True`` and a distributed sampler was already added, Lightning will not replace the existing one.
        For iterable-style datasets, we don't do this automatically. 
        """,
    )
    plugins = ListField(
        value=None,
        additions=[None],
        help="""
        Plugins allow modification of core behavior like ddp and amp, and enable custom lightning plugins.
        Default: ``None``.
        """,
    )
    sync_batchnorm = BoolField(
        value=False,
        help="""Synchronize batch norm layers between process groups/whole world. Default: ``False``.""",
    )
    reload_dataloaders_every_n_epochs = IntField(
        value=0,
        minimum=0,
        help=""" Set to a non-negative integer to reload dataloaders every n epochs. Default: ``0``. """,
    )

    submodule = SubModule(
        {},
        help=""" the callback modules configure""",
    )


@register("trainer", "lightning")
class LightningTrainer(object):
    """pytorch-lightning trainer"""

    def __init__(self, config: LightningTrainerConfig, rt_config: Dict):
        super().__init__()
        config_dict = config._to_dict(only_para=True)
        if config_dict["logger"]:
            config_dict["logger"] = TensorBoardLogger(
                save_dir=os.path.join(rt_config["log_dir"], rt_config["name"]),
                name="metrics",
                version="",
                default_hp_metric=False,
            )
        callbacks = []
        callback_configs = config._get_named_modules("callback")
        config_dict["enable_checkpointing"] = False
        for callback_config_name in callback_configs:
            config_dict.pop(f"@{callback_config_name}")
            callback_config = callback_configs[callback_config_name]
            if callback_config._module_name == "checkpoint":
                config_dict["enable_checkpointing"] = True
            callbacks.append(
                register.get(
                    "callback", register_module_name(callback_config._module_name)
                )(callback_config)(rt_config=rt_config)
            )

        config_dict["callbacks"] = callbacks
        config_dict["default_root_dir"] = rt_config["log_dir"]
        self.trainer = PLTrainer(**config_dict)

    @property
    def world_size(self):
        return self.trainer.world_size

    def fit(self, **inputs):
        """fit the model and datamodule to trainer

        Args:
            **inputs: dict of input, include "model", 'datamodule'

        Returns:
            Undefine

        """

        return self.trainer.fit(**inputs)

    def predict(self, **inputs):
        """fit the model and datamodule.predict_dataloader to predict

        Args:
            **inputs: dict of input, include "model", 'datamodule'

        Returns:
            predict list

        """
        return self.trainer.predict(**inputs)

    def test(self, **inputs):
        """fit the model and datamodule.test_dataloader to test

        Args:
            **inputs: dict of input, include "model", 'datamodule'

        Returns:
            Undefine

        """
        return self.trainer.test(**inputs)

    def validate(self, **inputs):
        """fit the model and datamodule.validation to validate

        Args:
            **inputs: dict of input, include "model", 'datamodule'

        Returns:
            Undefine

        """
        return self.trainer.validate(**inputs)



# Copyright the author(s) of DLK.
# Copyright spark-nlp-display.
#
# This original source code is copied from spark-nlp-display and licensed under the Apache license.

import json
import math
import os
import random
import re

import numpy as np
import pandas as pd
import svgwrite

here = os.path.abspath(os.path.dirname(__file__))
# overlap_hist = []
# y_hist_dict = {}
x_i_diff_dict = {}
x_o_diff_dict = {}

from intc import Base, DictField, IntField, StrField, dataclass


@dataclass
class SpanAnnotation(Base):
    annotatorType = StrField(
        value="",
        help="The type of the output of the annotator. Possible values are ``DOCUMENT, TOKEN, WORDPIECE, WORD_EMBEDDINGS, SENTENCE_EMBEDDINGS, CATEGORY, DATE, ENTITY, SENTIMENT, POS, CHUNK, NAMED_ENTITY, NEGEX, DEPENDENCY, LABELED_DEPENDENCY, LANGUAGE, KEYWORD, DUMMY``.",
    )
    begin = IntField(
        value=0, help="The index of the first character under this annotation."
    )
    end = IntField(
        value=0, help="The index of the last character under this annotation."
    )
    result = StrField(value="", help="The resulting string of the annotation.")
    metadata = DictField(value={}, help="Associated metadata for this annotation")


class RelationExtractionVisualizer:
    def __init__(self):
        with open(
            os.path.join(here, "label_colors/relations.json"), "r", encoding="utf-8"
        ) as f_:
            self.color_dict = json.load(f_)
        with open(
            os.path.join(here, "label_colors/ner.json"), "r", encoding="utf-8"
        ) as f_:
            self.entity_color_dict = json.load(f_)
            self.entity_color_dict = dict(
                (k.lower(), v) for k, v in self.entity_color_dict.items()
            )
            self.font_path = os.path.join(here, "fonts/Lucida_Console.ttf")
            self.main_font = "Lucida"

    def __get_color(self, l):
        r = lambda: random.randint(0, 200)
        return "#%02X%02X%02X" % (r(), r(), r())

    def __size(self, text):
        return ((len(text) + 1) * 9.7) - 5

    def __draw_line(
        self,
        dwg,
        s_x,
        s_y,
        e_x,
        e_y,
        d_type,
        color,
        show_relations,
        size_of_entity_label,
    ):
        eps = 0.0000001

        def get_bezier_coef(points):
            # since the formulas work given that we have n+1 points
            # then n must be this:
            n = len(points) - 1

            # build coefficents matrix
            C = 4 * np.identity(n)
            np.fill_diagonal(C[1:], 1)
            np.fill_diagonal(C[:, 1:], 1)
            C[0, 0] = 2
            C[n - 1, n - 1] = 7
            C[n - 1, n - 2] = 2

            # build points vector
            P = [2 * (2 * points[i] + points[i + 1]) for i in range(n)]
            P[0] = points[0] + 2 * points[1]
            P[n - 1] = 8 * points[n - 1] + points[n]

            # solve system, find a & b
            A = np.linalg.solve(C, P)
            B = [0] * n
            for i in range(n - 1):
                B[i] = 2 * points[i + 1] - A[i + 1]
            B[n - 1] = (A[n - 1] + points[n]) / 2

            return A, B

        # returns the general Bezier cubic formula given 4 control points
        def get_cubic(a, b, c, d):
            return (
                lambda t: np.power(1 - t, 3) * a
                + 3 * np.power(1 - t, 2) * t * b
                + 3 * (1 - t) * np.power(t, 2) * c
                + np.power(t, 3) * d
            )

        # return one cubic curve for each consecutive points
        def get_bezier_cubic(points):
            A, B = get_bezier_coef(points)
            return [
                get_cubic(points[i], A[i], B[i], points[i + 1])
                for i in range(len(points) - 1)
            ]

        # evalute each cubic curve on the range [0, 1] sliced in n points
        def evaluate_bezier(points, n):
            curves = get_bezier_cubic(points)
            return np.array([fun(t) for fun in curves for t in np.linspace(0, 1, n)])

        def draw_pointer(dwg, s_x, s_y, e_x, e_y):
            size = 5
            ratio = 1
            fullness1 = 2
            fullness2 = 3
            bx = e_x
            ax = s_x
            by = e_y
            ay = s_y
            abx = bx - ax
            aby = by - ay
            ab = np.sqrt(abx * abx + aby * aby) + eps

            cx = bx - size * abx / ab
            cy = by - size * aby / ab
            dx = cx + (by - cy) / ratio
            dy = cy + (cx - bx) / ratio
            ex = cx - (by - cy) / ratio
            ey = cy - (cx - bx) / ratio
            fx = (fullness1 * cx + bx) / fullness2
            fy = (fullness1 * cy + by) / fullness2

            text_place_y = s_y - (abs(s_y - e_y) / 2)
            """
            line = dwg.add(dwg.polyline(
                      [
                      (bx, by),    
                      (dx, dy),
                      (fx, fy),
                      (ex, ey),
                      (bx, by)
                      ],
                      stroke=color, stroke_width = "1", fill='none',))
            """
            line = dwg.add(
                dwg.polyline(
                    [(dx, dy), (bx, by), (ex, ey), (bx, by)],
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            return text_place_y

        unique_o_index = str(s_x) + str(s_y)
        unique_i_index = str(e_x) + str(e_y)
        if s_x > e_x:
            if unique_o_index in x_o_diff_dict:
                s_x -= 5
            else:
                s_x -= 10
                x_o_diff_dict[unique_o_index] = 5
            if s_y > e_y:
                e_x += size_of_entity_label
            elif s_y < e_y:
                s_x -= size_of_entity_label

            if unique_i_index in x_i_diff_dict:
                e_x += 5
            else:
                e_x += 10
                x_i_diff_dict[unique_i_index] = 5
        else:
            if unique_o_index in x_o_diff_dict:
                s_x += 5
            else:
                s_x += 10
                x_o_diff_dict[unique_o_index] = 5
            if s_y > e_y:
                e_x -= size_of_entity_label
            elif s_y < e_y:
                s_x += size_of_entity_label

            if unique_i_index in x_i_diff_dict:
                e_x -= 5
            else:
                e_x -= 10
                x_i_diff_dict[unique_i_index] = 5
        # this_y_vals = list(range(min(s_x,e_x), max(s_x,e_x)+1))
        # this_y_vals = [ str(s_y)+'|'+str(i) for i in this_y_vals]
        # common = set(this_y_vals) & set(overlap_hist)
        # overlap_hist.extend(this_y_vals)
        # if s_y not in y_hist_dict:
        #    y_hist_dict[s_y] = 20
        # if common:
        #    y_hist_dict[s_y] += 20
        # y_increase = y_hist_dict[s_y]
        angle = -1
        if s_y == e_y:
            angle = 0
            s_y -= 20
            e_y = s_y - 4  # 55

            text_place_y = s_y - 35

            pth = evaluate_bezier(
                np.array([[s_x, s_y], [(s_x + e_x) / 2.0, s_y - 40], [e_x, e_y]]), 50
            )
            dwg.add(
                dwg.polyline(
                    pth,
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            draw_pointer(dwg, (s_x + e_x) / 2.0, s_y - 50, e_x, e_y)
        elif s_y >= e_y:
            e_y += 15
            s_y -= 20
            text_place_y = s_y - (abs(s_y - e_y) / 2)

            pth = evaluate_bezier(
                np.array(
                    [
                        [s_x, s_y],
                        # [((3*s_x)+e_x)/4.0, (s_y+e_y)/2.0],
                        [(s_x + e_x) / 2.0, (s_y + e_y) / 2.0],
                        # [(s_x+(3*e_x))/4.0,(s_y+e_y)/2.0],
                        [e_x, e_y],
                    ]
                ),
                50,
            )
            dwg.add(
                dwg.polyline(
                    pth,
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            draw_pointer(dwg, s_x, s_y, e_x, e_y)

            """
            line = dwg.add(dwg.polyline(
                    [(s_x, s_y),(s_x, s_y-y_increase), (e_x, s_y-y_increase),
                    (e_x, e_y),    
                    (e_x+2, e_y),
                    (e_x, e_y-4),
                    (e_x-2, e_y),
                    (e_x, e_y)
                    ],
                    stroke=color, stroke_width = "2", fill='none',))
            """
        else:
            s_y += 15
            e_y -= 20
            text_place_y = s_y + (abs(s_y - e_y) / 2)

            pth = evaluate_bezier(
                np.array(
                    [
                        [s_x, s_y],
                        # [((3*s_x)+e_x)/4.0, (s_y+e_y)/2.0],
                        [(s_x + e_x) / 2.0, (s_y + e_y) / 2.0],
                        # [(s_x+(3*e_x))/4.0,(s_y+e_y)/2.0],
                        [e_x, e_y],
                    ]
                ),
                50,
            )
            dwg.add(
                dwg.polyline(
                    pth,
                    stroke=color,
                    stroke_width="1",
                    fill="none",
                )
            )
            draw_pointer(dwg, s_x, s_y, e_x, e_y)

        if show_relations:
            if angle == -1:
                angle = math.degrees(math.atan((s_y - e_y) / ((s_x - e_x) + eps)))
            rel_temp_size = self.__size(d_type) / 1.35
            rect_x, rect_y = (
                ((s_x + e_x) / 2.0) - (rel_temp_size / 2.0) - 3,
                text_place_y - 10,
            )
            rect_w, rect_h = (rel_temp_size + 3, 13)
            dwg.add(
                dwg.rect(
                    insert=(rect_x, rect_y),
                    rx=2,
                    ry=2,
                    size=(rect_w, rect_h),
                    fill="white",
                    stroke=color,
                    stroke_width="1",
                    transform=f"rotate({angle} {rect_x+rect_w/2} {rect_y+rect_h/2})",
                )
            )

            dwg.add(
                dwg.text(
                    d_type,
                    insert=(((s_x + e_x) / 2) - (rel_temp_size / 2.0), text_place_y),
                    fill=color,
                    font_size="12",
                    font_family="courier",
                    transform=f"rotate({angle} {rect_x+rect_w/2} {rect_y+rect_h/2})",
                )
            )

    def __gen_graph(self, rdf, selected_text, exclude_relations, show_relations):
        exclude_relations = [i.lower().strip() for i in exclude_relations]
        rdf = [i for i in rdf if i.result.lower().strip() not in exclude_relations]

        done_ent1 = {}
        done_ent2 = {}
        all_done = {}

        start_y = 75
        x_limit = 1000
        y_offset = 100
        # dwg = svgwrite.Drawing("temp.svg",profile='full', size = (x_limit, len(selected_text) * 1.1 + len(rdf)*20))

        begin_index = 0
        start_x = 10
        this_line = 0

        all_entities_index = set()
        all_entities_1_index = []
        basic_dict = {}
        relation_dict = {}
        for t in rdf:
            all_entities_index.add(int(t.metadata["entity1_begin"]))
            all_entities_index.add(int(t.metadata["entity2_begin"]))
            basic_dict[int(t.metadata["entity1_begin"])] = [
                t.metadata["entity1_begin"],
                t.metadata["entity1_end"],
                t.metadata["chunk1"],
                t.metadata["entity1"],
            ]

            basic_dict[int(t.metadata["entity2_begin"])] = [
                t.metadata["entity2_begin"],
                t.metadata["entity2_end"],
                t.metadata["chunk2"],
                t.metadata["entity2"],
            ]
            if t.metadata["entity1"].lower().strip() not in self.entity_color_dict:
                self.entity_color_dict[
                    t.metadata["entity1"].lower().strip()
                ] = self.__get_color(t.metadata["entity1"].lower().strip())
            if t.metadata["entity2"].lower().strip() not in self.entity_color_dict:
                self.entity_color_dict[
                    t.metadata["entity2"].lower().strip()
                ] = self.__get_color(t.metadata["entity2"].lower().strip())

            # all_entities_1_index.append(t[4]['entity1_begin'])
        all_entities_index = np.asarray(list(all_entities_index))
        all_entities_index = all_entities_index[np.argsort(all_entities_index)]
        dwg_rects, dwg_texts = [], []
        for ent_start_ind in all_entities_index:
            e_start_now, e_end_now, e_chunk_now, e_entity_now = basic_dict[
                ent_start_ind
            ]
            prev_text = selected_text[begin_index : int(e_start_now)]
            prev_text = re.sub(r"\s*(\n)+", r"\1", prev_text.strip(), re.MULTILINE)
            begin_index = int(e_end_now) + 1
            for line_num, line in enumerate(prev_text.split("\n")):
                if line_num != 0:
                    start_y += y_offset
                    start_x = 10
                    this_line = 0
                for word_ in line.split(" "):
                    this_size = self.__size(word_)
                    if (start_x + this_size + 10) >= x_limit:
                        start_y += y_offset
                        start_x = 10
                        this_line = 0
                    dwg_texts.append(
                        [
                            word_,
                            (start_x, start_y),
                            "#546c74",
                            "16",
                            self.main_font,
                            "font-weight:100",
                        ]
                    )
                    # dwg.add(dwg.text(word_, insert=(start_x, start_y ), fill='#546c77', font_size='16',
                    #                 font_family='Monaco', style='font-weight:lighter'))
                    start_x += this_size + 10

            this_size = self.__size(e_chunk_now)
            if (start_x + this_size + 10) >= x_limit:  # or this_line >= 2:
                start_y += y_offset
                start_x = 10
                this_line = 0

            # rectange chunk 1
            dwg_rects.append(
                [
                    (start_x - 3, start_y - 18),
                    (this_size, 25),
                    self.entity_color_dict[e_entity_now.lower().strip()],
                ]
            )
            # dwg.add(dwg.rect(insert=(start_x-3, start_y-18),rx=2,ry=2, size=(this_size,25), stroke=self.entity_color_dict[e_entity_now.lower()],
            # stroke_width='1', fill=self.entity_color_dict[e_entity_now.lower()], fill_opacity='0.2'))
            # chunk1
            dwg_texts.append(
                [
                    e_chunk_now,
                    (start_x, start_y),
                    "#546c74",
                    "16",
                    self.main_font,
                    "font-weight:100",
                ]
            )
            # dwg.add(dwg.text(e_chunk_now, insert=(start_x, start_y ), fill='#546c77', font_size='16',
            #                 font_family='Monaco', style='font-weight:lighter'))
            # entity 1
            central_point_x = start_x + (this_size / 2)
            temp_size = self.__size(e_entity_now) / 2.75
            dwg_texts.append(
                [
                    e_entity_now.upper(),
                    (central_point_x - temp_size, start_y + 20),
                    "#1f77b7",
                    "12",
                    self.main_font,
                    "font-weight:lighter",
                ]
            )
            # dwg.add(dwg.text(e_entity_now.upper(),
            #                insert=(central_point_x-temp_size, start_y+20),
            #                fill='#1f77b7', font_size='12', font_family='Monaco',
            #                style='font-weight:lighter'))

            all_done[int(e_start_now)] = [central_point_x, start_y, temp_size]
            start_x += this_size + 20
            this_line += 1

        prev_text = selected_text[begin_index:]
        prev_text = re.sub(r"\s*(\n)+", r"\1", prev_text.strip(), re.MULTILINE)
        for line_num, line in enumerate(prev_text.split("\n")):
            if line_num != 0:
                start_y += y_offset
                start_x = 10
            for word_ in line.split(" "):
                this_size = self.__size(word_)
                if (start_x + this_size) >= x_limit:
                    start_y += y_offset
                    start_x = 10
                dwg_texts.append(
                    [
                        word_,
                        (start_x, start_y),
                        "#546c77",
                        "16",
                        self.main_font,
                        "font-weight:100",
                    ]
                )
                # dwg.add(dwg.text(word_, insert=(start_x, start_y ), fill='#546c77', font_size='16',
                #                 font_family='Monaco', style='font-weight:lighter'))
                start_x += this_size + 10

        dwg = svgwrite.Drawing(
            "temp.svg", profile="full", size=(x_limit, start_y + y_offset)
        )
        dwg.embed_font(self.main_font, self.font_path)

        for crect_ in dwg_rects:
            dwg.add(
                dwg.rect(
                    insert=crect_[0],
                    rx=2,
                    ry=2,
                    size=crect_[1],
                    stroke=crect_[2],
                    stroke_width="1",
                    fill=crect_[2],
                    fill_opacity="0.2",
                )
            )

        for ctext_ in dwg_texts:
            dwg.add(
                dwg.text(
                    ctext_[0],
                    insert=ctext_[1],
                    fill=ctext_[2],
                    font_size=ctext_[3],
                    font_family=ctext_[4],
                    style=ctext_[5],
                )
            )

        relation_distances = []
        relation_coordinates = []
        for row in rdf:
            if row.result.lower().strip() not in self.color_dict:
                self.color_dict[row.result.lower().strip()] = self.__get_color(
                    row.result.lower().strip()
                )

            d_key2 = all_done[int(row.metadata["entity1_begin"])]
            d_key1 = all_done[int(row.metadata["entity2_begin"])]
            this_dist = abs(d_key2[0] - d_key1[0]) + abs(d_key2[1] - d_key1[1])
            relation_distances.append(this_dist)
            relation_coordinates.append((d_key2, d_key1, row.result))

        relation_distances = np.array(relation_distances)
        relation_coordinates = np.array(relation_coordinates, dtype=object)
        temp_ind = np.argsort(relation_distances)
        relation_distances = relation_distances[temp_ind]
        relation_coordinates = relation_coordinates[temp_ind]
        for row in relation_coordinates:
            # if int(row[0][1]) == int(row[1][1]):
            size_of_entity_label = int(row[1][2])
            self.__draw_line(
                dwg,
                int(row[0][0]),
                int(row[0][1]),
                int(row[1][0]),
                int(row[1][1]),
                row[2],
                self.color_dict[row[2].lower().strip()],
                show_relations,
                size_of_entity_label,
            )

        return dwg.tostring()

    def display(
        self,
        result,
        # relation_col,
        document_col="document",
        exclude_relations=["O"],
        show_relations=True,
        return_html=False,
        save_path=None,
    ):
        """Displays Relation Extraction visualization.
        Inputs:
        result -- A Dataframe or dictionary.
        relation_col -- Name of the column/key containing relationships.
        document_col -- Name of the column/key containing text document.
        exclude_relations -- list of relations that don't need to be displayed. Default: ["O"]
        show_relations -- Display relation types on arrows. Default: True
        return_html -- If true, returns raw html code instead of displaying. Default: False
        Output: Visualization
        """
        original_text = result["sentence"]
        entities = {}
        for entity in result["predict_entities_info"]:
            entity["chunk"] = original_text[entity["start"] : entity["end"]]
            entities[entity["entity_id"]] = entity
        relations = []
        for relation in result["predict_relations_info"]:
            relations.append(
                SpanAnnotation._from_dict(
                    {
                        "result": relation["labels"][0],
                        "metadata": {
                            "entity1": entities[relation["from"]]["labels"][0],
                            "entity1_begin": entities[relation["from"]]["start"],
                            "entity1_end": entities[relation["from"]]["end"],
                            "chunk1": entities[relation["from"]]["chunk"],
                            "entity2": entities[relation["to"]]["labels"][0],
                            "entity2_begin": entities[relation["to"]]["start"],
                            "entity2_end": entities[relation["to"]]["end"],
                            "chunk2": entities[relation["to"]]["chunk"],
                        },
                    }
                )
            )

        # original_text = result[document_col][0].result
        # res = result[relation_col]

        html_content = self.__gen_graph(
            relations, original_text, exclude_relations, show_relations
        )

        with open("/home/sun/Downloads/demo.html", "w") as f_:
            f_.write(html_content)


if __name__ == "__main__":
    vis = RelationExtractionVisualizer()
    result = {
        "sentence": "germany 's relations with russia were never likely to be as cozy under angela merkel as under her predecessor , gerhard schr\u00f6der , who adopted a 3-year-old russian girl and , on his 60th birthday , invited president vladimir v. putin home to celebrate .",
        "uuid": "16caf15f-e6d0-11ee-98a1-5bbc69525069",
        "predict_entities_info": [
            {
                "start": 0,
                "end": 7,
                "labels": ["person"],
                "sub_token_start": 1,
                "sub_token_end": 3,
                "entity_id": "cef224e8-e79a-11ee-ae44-18c04d299e80",
            },
            {
                "start": 26,
                "end": 32,
                "labels": ["ethnicity"],
                "sub_token_start": 8,
                "sub_token_end": 10,
                "entity_id": "cef22d44-e79a-11ee-ae44-18c04d299e80",
            },
            {
                "start": 112,
                "end": 128,
                "labels": ["people"],
                "sub_token_start": 30,
                "sub_token_end": 37,
                "entity_id": "cef240ea-e79a-11ee-ae44-18c04d299e80",
            },
            {
                "start": 156,
                "end": 163,
                "labels": ["people"],
                "sub_token_start": 47,
                "sub_token_end": 49,
                "entity_id": "cef24a9a-e79a-11ee-ae44-18c04d299e80",
            },
        ],
        "predict_relations_info": [
            {
                "from": "cef24a9a-e79a-11ee-ae44-18c04d299e80",
                "to": "cef22d44-e79a-11ee-ae44-18c04d299e80",
                "labels": ["geographic_distribution"],
            },
            {
                "from": "cef240ea-e79a-11ee-ae44-18c04d299e80",
                "to": "cef224e8-e79a-11ee-ae44-18c04d299e80",
                "labels": ["nationality"],
            },
        ],
    }
    dis = vis.display(result)
    # print(dis)
    # with open("/home/sun/Downloads/dis.html", "w") as f_:
    #     f_.write(dis)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict, List

import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from lightning.pytorch.callbacks import ModelCheckpoint

from dlk.utils.register import register


@cregister("callback", "checkpoint")
class CheckpointCallbackConfig(Base):
    """callback for save checkpoint on the condition"""

    monitor = StrField(value="???", help="the monitor metrics for checkpoint")
    filename = StrField(
        value=None,
        additions=[None],
        help="""
        checkpoint filename. Can contain named formatting options to be auto-filled.
        By default, filename is ``None`` and will be set to ``'{epoch}-{step}'``, where "epoch" and "step" match
        the number of finished epoch and optimizer steps respectively.
        """,
    )
    save_top_k = IntField(
        value=3,
        minimum=0,
        help="""
        if ``save_top_k == k``,
        the best k models according to the quantity monitored will be saved.
        if ``save_top_k == 0``, no models are saved.
        if ``save_top_k == -1``, all models are saved.
        Please note that the monitors are checked every ``every_n_epochs`` epochs.
        if ``save_top_k >= 2`` and the callback is called multiple
        times inside an epoch, the name of the saved file will be
        appended with a version count starting with ``v1``.
        """,
    )
    mode = StrField(
        value="max",
        options=["min", "max"],
        help="max or min for monitor metrics",
    )
    save_last = BoolField(value=True, help="awlways save last checkpoint")
    auto_insert_metric_name = BoolField(
        value=True, help="the save file name with or not metric name"
    )
    every_n_train_steps = IntField(
        value=None,
        minimum=0,
        additions=[None],
        help="Number of training steps between checkpoints, default use `every_n_epochs`, when set the `every_n_train_steps` will disable the `every_n_epochs`.",
    )
    train_time_interval = IntField(
        value=None,
        minimum=0,
        additions=[None],
        help="""
        Checkpoints are monitored at the specified time interval.
        For all practical purposes, this cannot be smaller than the amount
        of time it takes to process a single training batch. This is not
        guaranteed to execute at the exact time specified, but should be close.
        when set the `train_time_interval` will disable the `every_n_epochs` and `every_n_train_steps`.
        """,
    )
    every_n_epochs = IntField(
        value=1,
        minimum=0,
        help="Number of training epochs between checkpoints.",
    )
    save_on_train_epoch_end = BoolField(
        value=False,
        help="Whether to run checkpointing at the end of the training epoch. If this is False, then the check runs at the end of the validation.",
    )
    save_weights_only = BoolField(
        value=True,
        help="if ``True``, then only the model's weights will be saved. Otherwise, the optimizer states, lr-scheduler states, etc are added in the checkpoint too.",
    )
    verbose = BoolField(value=True, help="verbosity mode.")
    enable_version_counter = BoolField(value=False, help="enable version counter")


@register("callback", "checkpoint")
class CheckpointCallback(object):
    """callback for save checkpoint on the condition"""

    def __init__(self, config: CheckpointCallbackConfig):
        super().__init__()
        self.config = config._to_dict(only_para=True)
        if self.config["train_time_interval"] is not None:
            self.config["every_n_epochs"] = None
            self.config["every_n_train_steps"] = None

        if self.config["every_n_train_steps"] is not None:
            self.config["every_n_epochs"] = None

    def __call__(self, rt_config: Dict) -> ModelCheckpoint:
        """get the ModelCheckpoint object

        Args:
            rt_config: runtime config, include log_dir, and the checkpoint path name

        Returns:
            ModelCheckpoint object

        """
        dirpath = os.path.join(
            rt_config.get("log_dir", ""), rt_config.get("name", ""), "checkpoint"
        )
        return ModelCheckpoint(dirpath=dirpath, **self.config)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import dlk.nn.layer.decoder
import dlk.nn.layer.embedding
import dlk.nn.layer.encoder
import dlk.nn.layer.token_gen_decoder



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.display import Display, DisplayConfigBase
from dlk.utils.display.ner import NerVisualizer
from dlk.utils.register import register

logger = logging.getLogger(__name__)


class SeqLabDisplayConfig(DisplayConfigBase):
    """the seq_lab display"""

    title = StrField("Sequence Labeling Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the sentence to the `sentence input area` and click `submit` to get the entities info of this sentence",
        help="the help infomation of the demo",
    )

    render_height = IntField(value=500, help="the display height")
    render_width = IntField(value=900, help="the display width")
    render_type = StrField(value="html", help="the render type of the display")
    output_head = StrField(value="Sequence Label", help="the output head")

    class Input:
        sentence = StrField(
            "text",
            options=["text"],
            help="the input values name for seq lab, the value is the type",
        )

    input = NestField(value=Input, help="the input config")

    class Hold:
        entities_info = ListField([], help="the default value of entities_info")

    hold = NestField(value=Hold, help="the hold values")

    ignore_labels = ListField(
        value=["O", "X", "S", "E"],
        help="the ignore labels, if the entity label in this list, we will ignore this entity",
    )


cregister("display", "seq_lab")(SeqLabDisplayConfig)
cregister("display", "span_cls")(SeqLabDisplayConfig)


class SeqLabDisplay(Display):
    """sequence labeling display"""

    def __init__(self, config: SeqLabDisplayConfig):
        super(SeqLabDisplay, self).__init__(config)
        self.visual = NerVisualizer(config.ignore_labels)

    def display(self, result):
        return self.visual.display(result)


register("display", "seq_lab")(SeqLabDisplay)
register("display", "span_cls")(SeqLabDisplay)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
from typing import Callable, Dict, Union

import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer, normalizers, pre_tokenizers
from tokenizers.models import WordPiece

from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.tokenizer_util import (
    PreTokenizerFactory,
    TokenizerNormalizerFactory,
    TokenizerPostprocessorFactory,
)

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "fast_tokenizer")
class FastTokenizerConfig(BaseSubProcessorConfig):
    """FastTokenizer use hugingface tokenizers"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    predict_data_set = ListField(
        value=["predict"],
        suggestions=[["predict"], []],
        help="the data set should be processed for predict stage, only predict data set will be processed or none of the data set will be processed",
    )
    online_data_set = ListField(
        value=["online"],
        suggestions=[["online"], []],
        help="the data set should be processed for online stage, only online data set will be processed or none of the data set will be processed",
    )

    class InputMap:
        sentence = StrField(
            value="sentence",
            help="for single input, tokenize the 'sentence' column",
        )
        sentence_a = StrField(
            value="sentence_a",
            help="for pair inputs, tokenize the 'sentence_a' column",
        )
        sentence_b = StrField(
            value="sentence_b",
            help="for pair inputs, tokenize the 'sentence_b' column",
        )
        pretokenized_words = StrField(
            value="pretokenized_words", help="pretokenized word related to sentence"
        )
        pretokenized_words_a = StrField(
            value="pretokenized_words_a",
            help="pretokenized word b related to sentence_a",
        )
        pretokenized_words_b = StrField(
            value="pretokenized_words_b",
            help="pretokenized word b related to sentence_b",
        )
        pretokenized_word_offsets = StrField(
            value="pretokenized_word_offsets",
            help="pretokenized word offsets for fix offset",
        )
        pretokenized_word_offsets_a = StrField(
            value="pretokenized_word_offsets_a",
            help="pretokenized word offsets for fix offset",
        )
        pretokenized_word_offsets_b = StrField(
            value="pretokenized_word_offsets_b",
            help="pretokenized word offsets for fix offset",
        )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        tokens = StrField(value="tokens", help="the output tokens")
        ids = StrField(value="input_ids", help="the output input_ids")
        attention_mask = StrField(
            value="attention_mask", help="the output attention_mask"
        )
        type_ids = StrField(value="type_ids", help="the output type_ids")
        special_tokens_mask = StrField(
            value="special_tokens_mask", help="the output special_tokens_mask"
        )
        overflowing = StrField(value="overflowing", help="the output overflowing")
        offsets = StrField(value="offsets", help="the output offsets")
        word_ids = StrField(value="word_ids", help="the output word_ids")
        sequence_ids = StrField(value="sequence_ids", help="the output sequence_ids")

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    tokenizer_path = StrField(
        value=MISSING,
        help="the config path for the tokenizer, this is not effected by `meta_dir`",
    )

    class Truncation:
        direction = StrField(
            value="right",
            options=["right", "left"],
            help="the truncation direction",
        )
        stride = IntField(value=0, help="the stride for truncation")
        max_length = IntField(
            value=512, minimum=1, help="the max length for truncation"
        )
        strategy = StrField(
            value="longest_first",
            options=["longest_first", "only_first", "only_second"],
            help="the truncation strategy",
        )

    truncation = NestField(value=Truncation)
    normalizer = AnyField(
        value="default",
        suggestions=[
            "default",
            [
                "nfd",
                "lowercase",
                "strip_accents",
                {"some_processor_need_config": {}},
            ],
        ],
        help="the normalizer for the tokenizer, list of normalizer",
    )
    pre_tokenizer = AnyField(
        value="default",
        suggestions=[
            "default",
            [
                "whitespace",
                "whitespacesplit",
                "bytelevel",
                "bert",
                {"some_processor_need_config": {}},
            ],
        ],
        help="the pre tokenizer for the tokenizer, if not default, you can provide a list of pre tokenizers",
    )
    post_processor = AnyField(
        value="default",
        suggestions=["default", "bert", {"some_processor_need_config": {}}],
        help="the post processor for the tokenizer",
    )

    class ProcessData:
        is_pretokenized = BoolField(
            value=False, help="whether the input is pretokenized"
        )
        add_special_tokens = BoolField(value=True, help="whether add special tokens")

    process_data = NestField(value=ProcessData)
    expand_examples = BoolField(
        value=False,
        help="if the sequence is very long will split to multiple instance, whether expand the examples",
    )
    input_type = StrField(
        value="single",
        options=["single", "pair"],
        help="the input type of the tokenizer, single or pair",
    )
    fix_offset = BoolField(
        value=False, help="whether fix the offset for the pretokenized word"
    )


@register("subprocessor", "fast_tokenizer")
class FastTokenizer(BaseSubProcessor):
    """FastTokenizer use hugingface tokenizers

    Tokenizer the single $sentence
    Or tokenizer the pair $sentence_a, $sentence_b
    Generator $tokens, $input_ids, $type_ids, $special_tokens_mask, $offsets, $word_ids, $overflowing, $sequence_ids
    """

    def __init__(self, stage: str, config: FastTokenizerConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.config = config
        with open(self.config.tokenizer_path, "r", encoding="utf-8") as f:
            tokenizer_str = json.dumps(json.load(f))

        self.tokenizer = Tokenizer.from_str(tokenizer_str)
        pretokenizer_factory = PreTokenizerFactory(self.tokenizer)
        tokenizer_postprocessor_factory = TokenizerPostprocessorFactory(self.tokenizer)
        tokenizer_normalizer_factory = TokenizerNormalizerFactory(self.tokenizer)

        if not self.config.pre_tokenizer:
            self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence([])
        elif self.config.pre_tokenizer != "default":
            assert isinstance(self.config.pre_tokenizer, list)
            pre_tokenizers_list = []
            for one_pre_tokenizer in self.config.pre_tokenizer:
                pre_tokenizers_list.append(
                    self._get_processor(pretokenizer_factory, one_pre_tokenizer)
                )
            self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence(pre_tokenizers_list)

        if not self.config.post_processor:
            raise KeyError(
                "The tokenizer is not support disable default tokenizers post processor. (You can delete the config manually)"
            )
        elif self.config.post_processor != "default":
            self.tokenizer.post_processor = self._get_processor(
                tokenizer_postprocessor_factory, self.config.post_processor
            )

        if not self.config.normalizer:
            self.tokenizer.normalizer = normalizers.Sequence([])
        elif self.config.normalizer != "default":
            assert isinstance(self.config.normalizer, list)
            normalizers_list = []
            for one_normalizer in self.config.normalizer:
                normalizers_list.append(
                    self._get_processor(tokenizer_normalizer_factory, one_normalizer)
                )
            self.tokenizer.normalizer = normalizers.Sequence(normalizers_list)

        if self.config.truncation:
            self.tokenizer.enable_truncation(
                max_length=self.config.truncation.max_length,
                stride=self.config.truncation.stride,
                strategy=self.config.truncation.strategy,
                direction=self.config.truncation.direction,
            )

    def _get_processor(
        self,
        factory: Union[
            PreTokenizerFactory,
            TokenizerNormalizerFactory,
            TokenizerPostprocessorFactory,
        ],
        one_processor: Union[Dict, str],
    ):
        """return the processor in factory by the processor name and update the config of the processor if provide

        Args:
            factory: process factory
            one_processor: the processor info, it's name (and config)

        Returns:
            processor

        """
        if isinstance(one_processor, dict):
            assert len(one_processor) == 1
            process_name, process_config = list(one_processor.items())[0]
            return factory.get(process_name)(**process_config)
        else:
            assert isinstance(one_processor, str)
            return factory.get(one_processor)()

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """use self._tokenize tokenize the data

        Args:
            data: several data in dataframe

        Returns:
            updated dataframe

        """
        if self.config.input_type == "single":
            batch_encodes = self.tokenizer.encode_batch(
                data[self.config.input_map.pretokenized_words]
                if self.config.process_data.is_pretokenized
                else data[self.config.input_map.sentence],
                is_pretokenized=self.config.process_data.is_pretokenized,
                add_special_tokens=self.config.process_data.add_special_tokens,
            )
        else:  # pair
            sentence_as = (
                data[self.config.input_map.pretokenized_words_a]
                if self.config.process_data.is_pretokenized
                else data[self.config.input_map.sentence_a]
            )
            sentence_bs = (
                data[self.config.input_map.pretokenized_words_b]
                if self.config.process_data.is_pretokenized
                else data[self.config.input_map.sentence_b]
            )
            batch_encodes = self.tokenizer.encode_batch(
                [
                    (sentence_a, sentence_b)
                    for sentence_a, sentence_b in zip(sentence_as, sentence_bs)
                ],
                is_pretokenized=self.config.process_data.is_pretokenized,
                add_special_tokens=self.config.process_data.add_special_tokens,
            )
        if self.config.expand_examples:
            encodes_list = []
            for encode in batch_encodes:
                encodes_list.append([encode] + encode.overflowing)
            data["_tokenizer_encoders"] = encodes_list
            data = data.explode("_tokenizer_encoders", ignore_index=True)
        else:
            data["_tokenizer_encoders"] = batch_encodes
        (
            tokens_list,
            ids_list,
            attention_mask_list,
            type_ids_list,
            special_tokens_mask_list,
            offsets_list,
            word_ids_list,
            overflowing_list,
            sequence_ids_list,
        ) = ([], [], [], [], [], [], [], [], [])
        for encode in data["_tokenizer_encoders"]:
            tokens_list.append(encode.tokens)
            ids_list.append(encode.ids)
            attention_mask_list.append(encode.attention_mask)
            type_ids_list.append(encode.type_ids)
            special_tokens_mask_list.append(encode.special_tokens_mask)
            offsets_list.append(encode.offsets)
            word_ids_list.append(encode.word_ids)
            # overflowing_list.append(encode.overflowing)
            sequence_ids_list.append(encode.sequence_ids)
        output_map = self.config.output_map
        data[output_map.tokens] = tokens_list
        data[output_map.ids] = ids_list
        data[output_map.attention_mask] = attention_mask_list
        data[output_map.type_ids] = type_ids_list
        data[output_map.special_tokens_mask] = special_tokens_mask_list
        data[output_map.offsets] = offsets_list
        data[output_map.word_ids] = word_ids_list
        # data[output_map['overflowing']] = overflowing_list
        data[output_map.sequence_ids] = sequence_ids_list
        data.drop("_tokenizer_encoders", axis=1, inplace=True)

        if self.config.process_data.is_pretokenized and self.config.fix_offset:
            data[output_map.offsets] = data.apply(self._fix_offset, axis=1)
        return data

    def _fix_offset(self, one_line: pd.Series):
        """fix the pretokenizerd offset

        Args:
            one_line: a Series which contains the config.input_map.pretokenized_word_offsets, config.output_map.offsets, config.output_map.word_ids, configs.output_map.type_ids

        Returns:
            encode.tokens, encode.ids, encode.attention_mask, encode.type_ids, encode.special_tokens_mask, encode.offsets, encode.word_ids, encode.overflowing, encode.sequence_ids

        """
        fixed_offsets = []
        word_offsets = []
        if self.config.input_type == "single":
            word_offsets = [one_line[self.config.input_map.pretokenized_word_offsets]]
        else:  # pair
            word_offsets = [
                one_line[self.config.input_map.pretokenized_word_offsets_a],
                one_line[self.config.input_map.pretokenized_word_offsets_b],
            ]
        for offset, word_id, type_id in zip(
            one_line["offsets"], one_line["word_ids"], one_line["type_ids"]
        ):
            if offset == (0, 0):
                fixed_offsets.append(offset)
            else:
                fixed_offsets.append(
                    (
                        offset[0] + word_offsets[type_id][word_id][0],
                        offset[1] + word_offsets[type_id][word_id][0],
                    )
                )
        return fixed_offsets



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Set

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.nn.module.linear import Linear, LinearConfig
from dlk.utils.register import register


@cregister("decoder", "linear")
class DecoderLinearConfig(LinearConfig):
    """the linear decoder module"""

    class InputMap:
        embedding = StrField(value="embedding", help="the embedding")

    input_map = NestField(value=InputMap, help="the input map of the linear module")

    class OutputMap:
        logits = StrField(value="logits", help="the logits")

    output_map = NestField(value=OutputMap, help="the output map of the linear module")


@register("decoder", "linear")
class DecoderLinear(SimpleModule):
    """wrap for torch.nn.Linear"""

    def __init__(self, config: DecoderLinearConfig):
        super(DecoderLinear, self).__init__(config)
        self.linear = register.get("module", "linear")(config)
        self.config = config

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.linear.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """All step do this

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        inputs[self.config.output_map.logits] = self.linear(
            inputs[self.config.input_map.embedding]
        )
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.


import json
import os
from typing import Callable, Dict, List, Optional, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.nn.module.module_bart_decoder import BartDecoderWrap, BartDecoderWrapConfig
from dlk.utils.register import register


@cregister("token_gen_decoder", "bart_like_decoder")
class BartLikeDecoderConfig(BartDecoderWrapConfig):
    """the bart decoder module"""

    class InputMap:
        decoder_attention_mask = StrField(
            value="decoder_attention_mask", help="the decoder attention mask"
        )
        encoder_output_embedding = StrField(
            value="encoder_output_embedding", help="the encoder outputs"
        )
        decoder_head_mask = StrField(
            value="decoder_head_mask", help="the decoder head mask"
        )
        decoder_input_embedding = StrField(
            value="decoder_input_embedding", help="the input embedding"
        )

    bart_like_module_name = StrField(
        value="bart_decoder",
        help="the bart like decoder module name, like `bart_decoder` ",
    )
    input_map = NestField(value=InputMap, help="the input map of the linear module")

    class OutputMap:
        decoder_output_embedding = StrField(
            value="decoder_output_embedding", help="the output embedding"
        )
        decoder_past_cache = StrField(value="decoder_past_cache", help="the past cache")

    output_map = NestField(value=OutputMap, help="the output map of the linear module")


@register("token_gen_decoder", "bart_like_decoder")
class BartLikeDecoder(SimpleModule):
    """bart_like_decoder"""

    def __init__(self, config: BartLikeDecoderConfig, embedding: nn.Embedding):
        super(BartLikeDecoder, self).__init__(config)
        self.config = config

        self.bart_like_decoder = register.get("module", config.bart_like_module_name)(
            config=config, embedding=embedding
        )

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.bart_like_decoder.init_weight(method)

    def reorder_incremental_state(
        self,
        decoder_past_cache,
        new_order,
    ):
        if not decoder_past_cache:
            return decoder_past_cache
        return self.bart_like_decoder.reorder_incremental_state(
            decoder_past_cache, new_order
        )

    def forward(
        self, inputs: Dict[str, torch.Tensor], decoder_past_cache=None
    ) -> Dict[str, torch.Tensor]:
        """

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        module_inputs = {
            "decoder_attention_mask": inputs.get(
                self.config.input_map.decoder_attention_mask, None
            ),
            "encoder_outputs": inputs[self.config.input_map.encoder_output_embedding],
            "decoder_head_mask": inputs.get(
                self.config.input_map.decoder_head_mask, None
            ),
            "past_caches": decoder_past_cache,
            "inputs_embeds": inputs[self.config.input_map.decoder_input_embedding],
        }
        (
            hidden_states,
            next_cache,
            all_hidden_states,
            all_self_attns,
            all_cross_attentions,
        ) = self.bart_like_decoder(module_inputs)
        inputs[self.config.output_map.decoder_output_embedding] = hidden_states
        # rewrite the cache
        inputs[self.config.output_map.decoder_past_cache] = next_cache
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os

from setuptools import find_packages, setup


def write_version_py():
    with open(os.path.join("dlk", "version.txt")) as f:
        version = f.read().strip()

    # write version info to fairseq/version.py
    with open(os.path.join("dlk", "version.py"), "w") as f:
        f.write(f'__version__ = "{version}"\n')
    return version


version = write_version_py()


cmdclass = {}
extensions = []
if ("CUDA_HOME" in os.environ) and int(os.environ.get("BUILD_CUDA", "0")) == 1:
    try:
        from torch.utils import cpp_extension

        extensions = [
            cpp_extension.CppExtension(
                "dlk.ngram_repeat_block_cuda",
                sources=[
                    "dlk/cuda/ngram_repeat_block_cuda.cpp",
                    "dlk/cuda/ngram_repeat_block_cuda_kernel.cu",
                ],
            ),
        ]
        cmdclass = {"build_ext": cpp_extension.BuildExtension}
    except:
        pass

if "READTHEDOCS" in os.environ:
    # don't build extensions when generating docs
    extensions = []
    if "build_ext" in cmdclass:
        del cmdclass["build_ext"]


def package_files(directory):
    paths = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join(".", path, filename))
    return paths


added_files = []
added_files.extend(
    package_files(os.path.join("dlk", "utils", "display", "label_colors"))
)
added_files.extend(package_files(os.path.join("dlk", "utils", "display", "fonts")))
with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open("LICENSE", encoding="utf-8") as f:
    license = f.read()

pkgs = [p for p in find_packages() if p.startswith("dlk")]

setup(
    url="https://github.com/cstsunfu/dlk",
    description="dlk: Deep Learning Kit",
    long_description_content_type="text/markdown",
    version=version,
    ext_modules=extensions,
    cmdclass=cmdclass,
    package_data={"": added_files},
    license=license,
    python_requires=">=3.7",
    include_package_data=True,
    packages=pkgs,
    install_requires=[
        "intc>=0.1.0,<0.2",
        "hjson>=3.0.2",
        "pandas>=1.4",
        "pyarrow>=10.0.1",
        "fsspec>=2023.5,<2024.2",
        "torchmetrics>=1.2.1",
        "transformers>=4.24.0,<4.45",
    ],
    extras_require={
        "train": [
            "lightning>=2.1.0,<=2.2",
            "torch>=2.1,<=2.6",
            "tensorboard>=2.7.0,<3.0",
            "datasets>=2.17.0,<2.20",
            "torchmetrics>=1.2.1",
        ],
        "export": ["torch>=1.8.1,<2.5"],
        "process": [
            "datasets>=2.17.0,<2.20",
        ],
        "all": [
            "lightning>=2.1.0,<=2.2",
            "torch>=2.1,<=2.6",
            "tensorboard>=2.7.0,<3.0",
            "datasets>=2.17.0,<2.20",
            "streamlit>=1.32",
            "svgwrite>=1.4",
            "pyecharts>=2.0",
        ],
    },
)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
from typing import Dict

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from transformers.models.bert.configuration_bert import BertConfig
from transformers.models.bert.modeling_bert import BertModel

from dlk.utils.io import open
from dlk.utils.register import register

from . import Module


@cregister("module", "bert")
class BertWrapConfig:
    """the BERT config"""

    pretrained_model_path = StrField(value=MISSING, help="the pretrained model path")
    from_pretrain = BoolField(value=True, help="whether to load the pretrained model")
    freeze = BoolField(value=False, help="whether to freeze the model")
    dropout = FloatField(value=0.0, minimum=0.0, maximum=1.0, help="the dropout rate")


@register("module", "bert")
class BertWrap(Module):
    """Bert wrap"""

    def __init__(self, config: BertWrapConfig):
        super(BertWrap, self).__init__()
        self.config = config
        if os.path.isdir(self.config.pretrained_model_path):
            if os.path.exists(
                os.path.join(self.config.pretrained_model_path, "config.json")
            ):
                with open(
                    os.path.join(self.config.pretrained_model_path, "config.json"), "r"
                ) as f:
                    self.bert_config = BertConfig(**json.load(f))
            else:
                raise PermissionError(
                    f"config.json must in the dir {self.pretrained_model_path}"
                )
        else:
            if os.path.isfile(self.config.pretrained_model_path):
                try:
                    with open(self.config.pretrained_model_path, "r") as f:
                        self.bert_config = BertConfig(**json.load(f))
                except:
                    raise PermissionError(
                        f"You must provide the pretrained model dir or the config file path."
                    )

        self.bert = BertModel(self.bert_config, add_pooling_layer=True)
        self.dropout = nn.Dropout(float(self.config.dropout))

    def init_weight(self, method):
        """init the weight of model by 'bert.init_weight()' or from_pretrain

        Args:
            method: init method, no use for pretrained_transformers

        Returns:
            None

        """
        if self.config.from_pretrain:
            self.from_pretrained()
        else:
            self.bert.init_weights()

    def from_pretrained(self):
        """init the model from pretrained_model_path"""
        self.bert = BertModel.from_pretrained(self.config.pretrained_model_path)

    def forward(self, inputs: Dict):
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns:
            sequence_output, all_hidden_states, all_self_attentions

        """
        if self.config.freeze:
            with torch.no_grad():
                outputs = self.bert(
                    input_ids=inputs.get("input_ids", None),
                    attention_mask=inputs.get("attention_mask", None),
                    token_type_ids=inputs.get("token_type_ids", None),
                    position_ids=inputs.get("position_ids", None),
                    head_mask=inputs.get("head_mask", None),
                    inputs_embeds=inputs.get("inputs_embeds", None),
                    encoder_hidden_states=inputs.get("encoder_hidden_states", None),
                    encoder_attention_mask=inputs.get("encoder_attention_mask", None),
                    past_key_values=inputs.get("past_key_values", None),
                    use_cache=None,
                    output_attentions=True,
                    output_hidden_states=True,
                    return_dict=False,
                )
        else:
            outputs = self.bert(
                input_ids=inputs.get("input_ids", None),
                attention_mask=inputs.get("attention_mask", None),
                token_type_ids=inputs.get("token_type_ids", None),
                position_ids=inputs.get("position_ids", None),
                head_mask=inputs.get("head_mask", None),
                inputs_embeds=inputs.get("inputs_embeds", None),
                encoder_hidden_states=inputs.get("encoder_hidden_states", None),
                encoder_attention_mask=inputs.get("encoder_attention_mask", None),
                past_key_values=inputs.get("past_key_values", None),
                use_cache=None,
                output_attentions=True,
                output_hidden_states=True,
                return_dict=False,
            )
        assert (
            len(outputs) == 4
        ), f"Please check transformers version, the len(outputs) is 4 in version == 4.12|4.15, or check your config and remove the 'add_cross_attention'"
        sequence_output, all_hidden_states, all_self_attentions = (
            outputs[0],
            outputs[2],
            outputs[3],
        )
        sequence_output = self.dropout(sequence_output)
        return sequence_output, all_hidden_states, all_self_attentions



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


import math
from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register
from dlk.utils.token_generation_constraints import (
    ConstraintState,
    OrderedConstraintState,
    UnorderedConstraintState,
)

from . import Sample
from .beam_search import BeamSearch

constraint_id_map = {1: OrderedConstraintState, 2: UnorderedConstraintState}


@cregister("token_sample", "lexically_constrained_beam_search")
class LexicallyConstrainedBeamSearchConfig(Base):
    pass


@register("token_sample", "lexically_constrained_beam_search")
class LexicallyConstrainedBeamSearch(Sample):
    """Implements lexically constrained beam search as described in

        Fast Lexically Constrained Decoding with Dynamic Beam
        Allocation for Neural Machine Translation.  Post & Vilar,
        NAACL 2018.  https://www.aclweb.org/anthology/N18-1119/

    and

        Improved Lexically Constrained Decoding for Translation and
        Monolingual Rewriting. Hu et al, NAACL
        2019. https://www.aclweb.org/anthology/N19-1090/

    This is accomplished by maintaining, for each beam hypothesis, a
    ConstraintState object (see constraints.py) that tracks which
    constraints have been generated and using this information to
    shape the beam for each input sentence.
    """

    def __init__(self, tgt_dict, config: LexicallyConstrainedBeamSearchConfig):
        super().__init__(tgt_dict)
        self.vocab_size = len(tgt_dict)
        self.num_cands = 0
        self.supports_constraints = True

    def init_constraints(self, batch_constraints: Optional[Tensor], beam_size: int):
        self.constraint_states = []
        for constraint_tensor in batch_constraints:
            constraint_id = int(constraint_tensor[0].cpu())
            ConstraintClass = constraint_id_map[constraint_id]
            constraint_state = ConstraintClass.create(constraint_tensor[1:])
            self.constraint_states.append([constraint_state for i in range(beam_size)])

    def prune_sentences(self, batch_idxs: Tensor):
        self.constraint_states = [
            self.constraint_states[i] for i in batch_idxs.tolist()
        ]

    def update_constraints(self, active_hypos: Tensor):
        if self.constraint_states:
            batch_size = active_hypos.size(0)
            for sentid in range(batch_size):
                self.constraint_states[sentid] = [
                    self.constraint_states[sentid][i] for i in active_hypos[sentid]
                ]

    def step(
        self,
        step: int,
        lprobs: Tensor,
        scores: Optional[Tensor],
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        """
        A constrained step builds a large candidates list from the following:
        - the top 2 * {beam_size} items over the whole beam
        - for each item in the beam
          - the top {each_k} (default 1)
          - all next constraints
        We then compute the constrained state of each beam item, and assign
        stripe codes: 0 to the best in each bank, 1 to the 2nd-best, and so
        on. We then sort by (stripe, score), and truncate the list at
        2 * beam size.

        Args:
            step: the decoder step
            lprobs: (batch size, beam size, target vocab)
                the target-vocab distributions for each item in the beam.
        Return: A tuple of (scores, indices, beams, constraints) where:
            scores: (batch, output beam size)
                the scores of the chosen elements
            indices: (batch, output beam size)
                the target vocab indices of the chosen elements
            beams: (batch, output beam size)
                the 0-indexed hypothesis ids of the chosen elements
            constraints: (batch, output beam size)
                the new constraint states
        """
        each_k = 1
        device = lprobs.device

        batch_size, beam_size, vocab_size = lprobs.size()

        self.num_cands = min(
            # Just take the k-best. We'll get another k from the 1-best from each
            # row, plus more from the constraints
            beam_size * 2,
            lprobs.view(batch_size, -1).size(1) - 1,  # -1 so we never select pad
        )

        # STEP 0: Preliminary. Prevent EOS for unfinished hyps across all batch items
        constraint_states = self.constraint_states
        if constraint_states and step > 0:
            not_finished_indices = []
            for sentno, sent_constraints in enumerate(constraint_states):
                for beamno, state in enumerate(sent_constraints):
                    index = sentno * beam_size + beamno
                    if not state.finished:
                        not_finished_indices.append(index)
            not_finished_indices = torch.tensor(not_finished_indices)
            if not_finished_indices.numel() > 0:
                lprobs.view(batch_size * beam_size, -1)[
                    not_finished_indices, self.eos
                ] = -math.inf

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam entry for each batch item
            lprobs = lprobs[:, ::beam_size, :].contiguous()
        else:
            # make probs contain cumulative scores for each hypothesis
            assert scores is not None
            lprobs = lprobs + scores[:, :, step - 1].unsqueeze(-1)

        top_prediction = torch.topk(
            lprobs.view(batch_size, -1),
            self.num_cands,
        )
        scores_buf, indices_buf = top_prediction
        # Project back into relative indices and beams
        beams_buf = indices_buf // vocab_size
        indices_buf = indices_buf.fmod(vocab_size)

        # Short circuit if there are no constraints in this batch
        if not constraint_states:
            return scores_buf, indices_buf, beams_buf

        # STEP 1: get top-1 from each hypothesis across all sentences in the batch
        if step > 0:
            top_scores, top_indices = torch.topk(
                lprobs.view(batch_size * beam_size, -1),
                k=each_k,
                dim=1,
            )
            top_scores = top_scores.view(batch_size, -1)
            top_indices = top_indices.view(batch_size, -1)
            scores_buf = torch.cat((scores_buf, top_scores), dim=1)
            indices_buf = torch.cat((indices_buf, top_indices), dim=1)
            new_beams = torch.arange(0, beam_size, device=device).repeat(batch_size, 1)
            beams_buf = torch.cat((beams_buf, new_beams), dim=1)

        # Now, process sentences in the batch one by one.
        new_scores_buf = torch.zeros((batch_size, 2 * beam_size), device=device)
        new_indices_buf = torch.zeros((batch_size, 2 * beam_size), device=device).long()
        new_beams_buf = torch.zeros((batch_size, 2 * beam_size), device=device).long()
        for sentno, states in enumerate(constraint_states):
            scores, indices, beams, new_states = self.step_sentence(
                step,
                sentno,
                lprobs[sentno],
                constraint_states[sentno],
                beams_buf[sentno].clone(),
                indices_buf[sentno].clone(),
                scores_buf[sentno].clone(),
            )
            new_scores_buf[sentno] = scores
            new_indices_buf[sentno] = indices
            new_beams_buf[sentno] = beams
            self.constraint_states[sentno] = new_states

        return new_scores_buf, new_indices_buf, new_beams_buf

    def step_sentence(
        self,
        step: int,
        sentno: int,
        lprobs: Tensor,
        constraint_states: List[List[ConstraintState]],
        beams_buf: Tensor,
        indices_buf: Tensor,
        scores_buf: Tensor,
    ):
        """Does per-sentence processing. Adds all constraints for each
        hypothesis to the list of candidates; then removes duplicates,
        sorts, and dynamically stripes across the banks. All tensor inputs
        are collapsed to those pertaining to a single input sentence.
        """
        device = lprobs.device

        # STEP 2: Add all constraints for each beam item
        for beamno, state in enumerate(constraint_states):
            next_tokens = torch.tensor(list(state.next_tokens()), device=device).long()
            if next_tokens.numel() != 0:
                indices_buf = torch.cat((indices_buf, next_tokens))
                next_beams = (
                    torch.tensor(beamno, device=device)
                    .repeat(next_tokens.size(0))
                    .long()
                )
                beams_buf = torch.cat((beams_buf, next_beams))
                next_values = lprobs[beamno].take(next_tokens.view(-1))
                scores_buf = torch.cat((scores_buf, next_values))

            # At the 0th time step, there is just one beam item
            if step == 0:
                break

        # STEP 3: Compute the "bank" for each candidate. This is the
        # number of constraints it's generated. We need this so that
        # we can do round-robin allocation of the beam across these
        # banks. If C is the number of constraints, we select the best
        # item in bank C, then the best in bank C-1, etc, followed by
        # the 2nd-best in bank C, the 2nd-best in bank C-1, etc, and so
        # on, until the maximum beam size. We accomplish this by
        # creating a sort key and striping across the banks.

        # Compute the new states for all candidates
        cands_size = indices_buf.size(0)
        constraint_states = [
            constraint_states[beams_buf[i]].advance(indices_buf[i])
            for i in range(cands_size)
        ]

        banks = torch.tensor([state.bank for state in constraint_states], device=device)

        # STEP 4: Sort
        num_constraint_tokens = len(state.tokens)

        # Sort by keys (bank, score) (i.e., sort banks together, and scores
        # within banks). AFAIK pytorch doesn't support either stable sort or
        # multi-key sorting, so we have to hack this.
        MAX_SCORE = -100
        sort_key = (num_constraint_tokens - banks) * MAX_SCORE + scores_buf
        sort_values, sort_indices = sort_key.sort(dim=0, descending=True)
        scores_buf = scores_buf[sort_indices]
        indices_buf = indices_buf[sort_indices]
        beams_buf = beams_buf[sort_indices]
        banks = banks[sort_indices]

        # Sort the constraints to follow suit
        constraint_states = [constraint_states[i] for i in sort_indices]

        # STEP 5: Remove duplicates. The topk calls (overall and
        # per-row) plus the per-row generation of constraints will
        # produce duplicates. Here we remove them.

        def roll(t):
            """Rolls a 1d tensor left by 1.

            [0, 1, 2, 3, 4] becomes [4, 0, 1, 2, 3]
            """
            return torch.cat((t[-1].unsqueeze(0), t[0:-1]), dim=0)

        # We map candidates (beam, token_id) to a single dimension.
        # This is then shifted by 1. We can then easily identify
        # duplicates and create a mask that identifies unique
        # extensions.
        uniques_mask = beams_buf * (self.vocab_size + 1) + indices_buf
        uniques_mask = roll(uniques_mask) != uniques_mask

        # Use the mask to pare down the data structures
        scores_buf = torch.masked_select(scores_buf, uniques_mask)
        indices_buf = torch.masked_select(indices_buf, uniques_mask)
        beams_buf = torch.masked_select(beams_buf, uniques_mask)
        banks = torch.masked_select(banks, uniques_mask)
        i = 1
        for mask in uniques_mask[1:]:
            if not mask:
                constraint_states.pop(i)
            i += mask

        # STEP 6: Assign IDs round-robin across banks, sort, and
        # truncate. Now that the candidates are sorted by (bank,
        # score) and uniqed, we dynamically allocate the {beam_size}
        # beam by striping across the candidates. These stripes will
        # be used as sort keys to do round-robin selection. This is
        # accomplished in a single pass with offsets. Sorting by
        # highest-banks (furthest-along hypotheses) first ensures
        # progress through the constraints.
        #
        # e.g., BANKS: 3 3 3 2 2 2 2 1 1 1 0 0
        # OLD STRIPES: 0 1 2 0 1 2 3 0 1 2 0 1
        # NEW STRIPES: 0 1+4 2+8 0+1 1+5 2+9 3+11 0+2 1+6 2+10 0+3 1+7
        #            = 0 5 10 1 6 11 13 2 7 12 3 8
        #
        # Sorting by this then gives the following banks:
        #
        #             3 2 1 0 3 2 1 0 3 2 1 2
        #
        # We'll take the top {beam_size} of these.
        stripe_offsets = [offset * (len(banks) + 1) for offset in range(len(banks) + 1)]
        stripes = torch.zeros_like(banks)
        cur_bank_count = -1
        cur_bank = banks[0]
        for i, bank in enumerate(banks):
            if bank != cur_bank:
                cur_bank_count = 0
                cur_bank = bank
            else:
                cur_bank_count += 1
            stripes[i] = num_constraint_tokens - bank + stripe_offsets[cur_bank_count]

        # STEP 7: Sort by the stripes values
        sort_values, sort_indices = stripes.sort(dim=0)
        scores_buf = scores_buf[sort_indices]
        indices_buf = indices_buf[sort_indices]
        beams_buf = beams_buf[sort_indices]
        constraint_states = [constraint_states[i] for i in sort_indices]

        # STEP 8: Truncate to the candidates size!
        scores_buf = scores_buf[: self.num_cands]
        indices_buf = indices_buf[: self.num_cands]
        beams_buf = beams_buf[: self.num_cands]

        return scores_buf, indices_buf, beams_buf, constraint_states



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional

import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.module import Module
from dlk.nn.utils.rope import RoFormerSinusoidalPositionalEmbedding
from dlk.utils.register import register


@cregister("module", "bilinear")
class BiLinearConfig:
    """the bilinear module config"""

    input_size = IntField(
        value=MISSING,
        minimum=0,
        help="the input size of the bilinear module",
    )
    hidden_size = IntField(
        value=0,
        minimum=0,
        help="the hidden size of the bilinear module, if set to 0, will set the hidden size to the input size",
    )
    output_size = IntField(
        value=MISSING,
        minimum=0,
        help="the output size of the bilinear module",
    )
    dropout = FloatField(
        value=0.0,
        minimum=0.0,
        maximum=1.0,
        help="the dropout rate of the bilinear module",
    )
    max_seq_len = IntField(
        value=1024,
        minimum=0,
        help="the max sequence length of the bilinear module",
    )
    relation_position = BoolField(
        value=False, help="whether to use the relative position"
    )
    active = StrField(
        value="leaky_relu",
        options=["none", "leaky_relu", "relu", "gelu", "glu", "selu", "celu", "elu"],
        help="the activation function",
    )


active_map = {
    "leaky_relu": nn.LeakyReLU,
    "relu": nn.ReLU,
    "gelu": nn.GELU,
    "glu": nn.GLU,
    "selu": nn.SELU,
    "celu": nn.CELU,
    "elu": nn.ELU,
}


@register("module", "bilinear")
class BiLinear(Module):
    """"""

    def __init__(self, config: BiLinearConfig):
        super(BiLinear, self).__init__()
        self.config = config
        if self.config.hidden_size == 0:
            self.config.hidden_size = self.config.input_size
        self.linear_a = nn.Linear(self.config.input_size, self.config.hidden_size)
        self.linear_b = nn.Linear(self.config.input_size, self.config.hidden_size)
        if self.config.relation_position:
            self.embed_positions = RoFormerSinusoidalPositionalEmbedding(
                self.config.max_seq_len, self.config.hidden_size
            )
        self.dropout = nn.Dropout(p=self.config.dropout)
        if config.active != "none":
            self.active = active_map[config.active]()
        else:
            self.active = lambda x: x

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.linear_a.apply(method)
        self.linear_b.apply(method)

    def forward(self, embedding: torch.Tensor) -> torch.Tensor:
        """do forward on a mini batch

        Args:
            embedding: a mini batch embedding, shape==(batch_size, input_a_len, input_size)

        Returns:
            input_a x bilinear x input_b, shape==(batch_size, input_a_len, input_b_len, output_size)

        """

        input_a = self.dropout(self.active(embedding))
        input_b = self.dropout(self.active(embedding))
        input_a = self.linear_a(input_a)
        input_b = self.linear_b(input_b)

        if self.config.relation_position:
            sinusoidal_pos = self.embed_positions(input_a.shape[:-1])[None, :, :]
            input_a, input_b = self.embed_positions.apply_rotary_position_embeddings(
                sinusoidal_pos, input_a, input_b
            )

        return output



# Copyright the author(s) of DLK.
#
# There are many code borrowed from fairseq.

""" Wrapper for ngram_repeat_block cuda extension """
import math
import warnings
from typing import List

import torch
from torch import nn

try:
    from dlk import ngram_repeat_block_cuda

    EXTENSION_BUILT = True
except ImportError:
    EXTENSION_BUILT = False


def is_cuda_extension_usable() -> bool:
    """Check whether ngram_repeat_block_cuda is built properly"""
    if not EXTENSION_BUILT or not torch.cuda.is_available():
        return False
    bsz = 2
    tokens = torch.tensor([[4, 4, 3, 2], [1, 2, 3, 4]], dtype=torch.long, device="cuda")
    lprobs = torch.rand((8, 12), device="cuda")
    try:
        outputs = ngram_repeat_block_cuda.forward(tokens, lprobs, bsz, 3, 4, 3)
        outputs = outputs + 4  # This line breaks if the extension is built incorrectly.
        return True
    except RuntimeError:
        warnings.warn(
            "NGramRepeatBlock extension must be rebuilt."
            'Run TORCH_CUDA_ARCH_LIST="6.0;6.1;7.0" python setup.py build_ext --inplace'
        )
        return False


class NGramRepeatBlock(nn.Module):
    """Wrapper class for calling ngram_repeat_block cuda extension"""

    def __init__(self, no_repeat_ngram_size: int, use_extension: bool = False):
        super().__init__()
        self.use_extension = is_cuda_extension_usable() if use_extension else False
        self.no_repeat_ngram_size = no_repeat_ngram_size

    @torch.jit.unused
    def call_cuda_extension(
        self,
        tokens,
        lprobs,
        bsz: int,
        beam_size: int,
        step: int,
    ):
        return ngram_repeat_block_cuda.forward(
            tokens, lprobs, bsz, step, beam_size, self.no_repeat_ngram_size
        )

    def reset_parameters(self):
        pass

    def forward(self, tokens, lprobs, bsz: int, beam_size: int, step: int):
        """
        Args:
            tokens(Tensor): Input tokens(Bsz*beam, seq_len)
            lprobs(Tensor): likelihood probability,
            Expected to be updated in place.(Bsz*beam, vocab_size)
            bsz(int): batch size
            step(int): current step
            beam_size(int): beam size
            no_repeat_ngram_size(int): Ngram size
        """
        msg = f"expected {bsz *beam_size} got"
        assert tokens.size(0) == bsz * beam_size, f"{msg} {tokens.size(0)}"
        assert lprobs.size(0) == bsz * beam_size, f"{msg} {lprobs.size(0)}"

        if self.use_extension:
            return self.call_cuda_extension(tokens, lprobs, bsz, beam_size, step)
        return self._no_repeat_ngram(
            tokens,
            lprobs,
            bsz,
            beam_size,
            step,
        )

    def _no_repeat_ngram(self, tokens, lprobs, bsz: int, beam_size: int, step: int):
        """For each hypothesis generate a list of previous ngrams and set associated lprobs to -inf"""
        banned_tokens = [
            torch.jit.annotate(List[int], []) for bbsz_idx in range(bsz * beam_size)
        ]
        if step + 2 - self.no_repeat_ngram_size >= 0:
            cpu_tokens: List[List[int]] = tokens.cpu().tolist()
            check_start_pos = step + 2 - self.no_repeat_ngram_size
            for bbsz_idx in range(bsz * beam_size):
                ngram_to_check = cpu_tokens[bbsz_idx][
                    -(self.no_repeat_ngram_size - 1) :
                ]
                for i in range(check_start_pos):
                    if (
                        ngram_to_check
                        == cpu_tokens[bbsz_idx][i : i + self.no_repeat_ngram_size - 1]
                    ):
                        banned_tokens[bbsz_idx].append(
                            cpu_tokens[bbsz_idx][i + self.no_repeat_ngram_size - 1]
                        )
        for bbsz_idx in range(bsz * beam_size):
            lprobs[bbsz_idx][
                torch.tensor(banned_tokens[bbsz_idx], dtype=torch.int64)
            ] = torch.tensor(-math.inf).to(lprobs)
        return lprobs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
    dataclass,
)

from dlk.utils.import_module import import_module_dir


@dataclass
class AdvMethodConfig(Base):
    """the base Adv method"""

    use_adv = BoolField(
        value=True,
        help="whether use adv method, effective on the imodel, if the config include advmethod and not set use_adv to False, default is True",
    )


class AdvMethod(object):
    """Save fgm decided by config"""

    def __init__(self, model: nn.Module, config):
        super().__init__()

    def training_step(self, imodel, batch: Dict[str, torch.Tensor], batch_idx: int):
        """do training_step on a mini batch

        Args:
            imodel: imodel instance
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns:
            the outputs

        """
        raise NotImplementedError


# automatically import any Python files in the models directory
adv_method_dir = os.path.dirname(__file__)
import_module_dir(adv_method_dir, "dlk.adv_method")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Callable, Dict

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.optim import Optimizer

from dlk.utils.register import register

from . import BaseOptimizer, BaseOptimizerConfig


@cregister("optimizer", "adan")
class AdanOptimizerConfig(BaseOptimizerConfig):
    """
    the adan optimizer, ADAptive Nesterov momentum algorithm
    Adan: Adaptive Nesterov Momentum Algorithm for Faster Optimizing Deep Models[J]. arXiv preprint arXiv:2208.06677, 2022.
         https://arxiv.org/abs/2208.06677
    Implementation adapted from https://github.com/sail-sg/Adan
    """

    lr = FloatField(value=1e-3, minimum=0.0, help="the learning rate of optimizer")
    betas = ListField(
        value=[0.98, 0.92, 0.99],
        help="beta 1-2-3 as described in paper - author says most sensitive to beta3 tuning",
    )
    eps = FloatField(value=1e-8, minimum=1e-20, help="the epsilon of optimizer")
    weight_decay = FloatField(
        value=2e-2, minimum=0.0, help="the weight decay of the optimizer"
    )


class Adan(Optimizer):
    """
    Implements a pytorch variant of Adan
    Adan was proposed in
    Adan: Adaptive Nesterov Momentum Algorithm for Faster Optimizing Deep Models[J]. arXiv preprint arXiv:2208.06677, 2022.
    https://arxiv.org/abs/2208.06677
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining parameter groups.
        lr (float, optional): learning rate. (default: 1e-3)
        betas (Tuple[float, float, flot], optional): coefficients used for computing
            running averages of gradient and its norm. (default: (0.98, 0.92, 0.99))
        eps (float, optional): term added to the denominator to improve
            numerical stability. (default: 1e-8)
        weight_decay (float, optional): decoupled weight decay (L2 penalty) (default: 0)
        no_prox (bool): how to perform the decoupled weight decay (default: False)
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.98, 0.92, 0.99),
        eps=1e-8,
        weight_decay=0.0,
        no_prox=False,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= betas[2] < 1.0:
            raise ValueError("Invalid beta parameter at index 2: {}".format(betas[2]))
        defaults = dict(
            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, no_prox=no_prox
        )
        super(Adan, self).__init__(params, defaults)

    @torch.no_grad()
    def restart_opt(self):
        for group in self.param_groups:
            group["step"] = 0
            for p in group["params"]:
                if p.requires_grad:
                    state = self.state[p]
                    # State initialization

                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p)
                    # Exponential moving average of gradient difference
                    state["exp_avg_diff"] = torch.zeros_like(p)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            beta1, beta2, beta3 = group["betas"]
            # assume same step across group now to simplify things
            # per parameter step can be easily support by making it tensor, or pass list into kernel
            if "step" in group:
                group["step"] += 1
            else:
                group["step"] = 1

            bias_correction1 = 1.0 - beta1 ** group["step"]
            bias_correction2 = 1.0 - beta2 ** group["step"]
            bias_correction3 = 1.0 - beta3 ** group["step"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad

                state = self.state[p]
                if len(state) == 0:
                    state["exp_avg"] = torch.zeros_like(p)
                    state["exp_avg_diff"] = torch.zeros_like(p)
                    state["exp_avg_sq"] = torch.zeros_like(p)
                    state["pre_grad"] = grad.clone()

                exp_avg, exp_avg_sq, exp_avg_diff = (
                    state["exp_avg"],
                    state["exp_avg_diff"],
                    state["exp_avg_sq"],
                )
                grad_diff = grad - state["pre_grad"]

                exp_avg.lerp_(grad, 1.0 - beta1)  # m_t
                exp_avg_diff.lerp_(grad_diff, 1.0 - beta2)  # diff_t (v)
                update = grad + beta2 * grad_diff
                exp_avg_sq.mul_(beta3).addcmul_(
                    update, update, value=1.0 - beta3
                )  # n_t

                denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction3)).add_(
                    group["eps"]
                )
                update = (
                    exp_avg / bias_correction1 + beta2 * exp_avg_diff / bias_correction2
                ).div_(denom)
                if group["no_prox"]:
                    p.data.mul_(1 - group["lr"] * group["weight_decay"])
                    p.add_(update, alpha=-group["lr"])
                else:
                    p.add_(update, alpha=-group["lr"])
                    p.data.div_(1 + group["lr"] * group["weight_decay"])

                state["pre_grad"].copy_(grad)

        return loss


@register("optimizer", "adan")
class AdanOptimizer(BaseOptimizer):
    """wrap for adan optimizer"""

    def __init__(self, model: nn.Module, config: AdanOptimizerConfig):
        super(AdanOptimizer, self).__init__(model, config, Adan)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict, List, Type, Union

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader, Dataset

from dlk.data.datamodule import IBaseDataModule
from dlk.data.dataset.default import DefaultDataset, DefaultDatasetConfig
from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


@cregister("datamodule", "default")
class BasicDatamoduleConfig(Base):
    """the default datamodule"""

    pin_memory = AnyField(
        value=None,
        options=[None, True, False],
        help="whether to use pin_memory, if set to None, we will check the gpu is available",
    )
    num_workers = IntField(
        value=1,
        minimum=-1,
        help="the number of workers for dataloader, if set to -1, we will use os.cpu_count() to get the cpu count",
    )
    shuffle = BoolField(value=True, help="whether shuffle the training data")
    train_batch_size = IntField(
        value=32, minimum=1, help="the batch size of train dataloader"
    )
    predict_batch_size = IntField(
        value=32, minimum=1, help="the batch size of predict dataloader"
    )
    online_batch_size = IntField(
        value=1, minimum=1, help="the batch size of online dataloader"
    )

    submodule = SubModule({}, suggestions=["dataset", "data_collate"])


@register("datamodule", "default")
class BasicDatamodule(IBaseDataModule):
    """Basic and General DataModule"""

    def __init__(
        self, config: BasicDatamoduleConfig, data: Dict[str, Any], rt_config: Dict
    ):
        super().__init__()

        self.config = config
        self.rt_config = rt_config
        self._online_key_type_pairs = None
        self.train_data = None
        self.valid_data = None
        self.test_data = None
        self.predict_data = None
        self.dataset_config: DefaultDatasetConfig = config.submodule.dataset
        self.dataset_creator: Type[DefaultDataset] = register.get(
            "dataset", register_module_name(self.dataset_config._module_name)
        )
        if self.config.num_workers == -1:
            import os

            self.config.num_workers = os.cpu_count()
        if "train" in data:
            self.train_data = self.dataset_creator(
                self.dataset_config, data["train"], rt_config
            )
        if "test" in data:
            self.test_data = self.dataset_creator(
                self.dataset_config, data["test"], rt_config
            )
        if "valid" in data:
            self.valid_data = self.dataset_creator(
                self.dataset_config, data["valid"], rt_config
            )
        if "predict" in data:
            self.predict_data = self.dataset_creator(
                self.dataset_config, data["predict"], rt_config
            )
        data_collate_config = config.submodule.data_collate
        self.collate_fn = register.get(
            "data_collate", register_module_name(data_collate_config._module_name)
        )(data_collate_config)

    def train_dataloader(self):
        """get the train set dataloader"""
        if not self.train_data:
            return None
        return DataLoader(
            self.train_data,
            batch_size=self.config.train_batch_size,
            collate_fn=self.collate_fn,
            pin_memory=self.config.pin_memory,
            shuffle=self.config.shuffle,
            num_workers=self.config.num_workers,
        )

    def predict_dataloader(self):
        """get the predict set dataloader"""
        if not self.predict_data:
            return None
        return DataLoader(
            self.predict_data,
            batch_size=self.config.predict_batch_size,
            collate_fn=self.collate_fn,
            pin_memory=self.config.pin_memory,
            shuffle=False,
            num_workers=self.config.num_workers,
        )

    def val_dataloader(self):
        """get the validation set dataloader"""
        if not self.valid_data:
            return None
        return DataLoader(
            self.valid_data,
            batch_size=self.config.predict_batch_size,
            collate_fn=self.collate_fn,
            pin_memory=self.config.pin_memory,
            shuffle=False,
            num_workers=self.config.num_workers,
        )

    def test_dataloader(self):
        """get the test set dataloader"""
        if not self.test_data:
            return None
        return DataLoader(
            self.test_data,
            batch_size=self.config.predict_batch_size,
            collate_fn=self.collate_fn,
            pin_memory=self.config.pin_memory,
            shuffle=False,
            num_workers=self.config.num_workers,
        )

    def online_dataloader(self, data):
        """get the online dataloader"""
        if not self._online_key_type_pairs:
            self._online_key_type_pairs = self.dataset_creator.real_key_type_pairs(
                self.dataset_config.key_type_pairs, data
            )
        dataset = self.dataset_creator(
            self.dataset_config, data, self.rt_config, self._online_key_type_pairs
        )
        return DataLoader(
            dataset,
            batch_size=self.config.predict_batch_size,
            collate_fn=self.collate_fn,
            pin_memory=self.config.pin_memory,
            shuffle=False,
            num_workers=1,
        )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import io
import json
import logging
from typing import Any, Callable, Dict, List, Union
from uuid import uuid4

import hjson
import numpy as np
import pandas as pd
import streamlit as st
import streamlit.components.v1 as components
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    Parser,
    StrField,
    SubModule,
    asdict,
    cregister,
    init_config,
)
from PIL import Image

from dlk.display import Display
from dlk.online import OnlinePredict
from dlk.preprocess import PreProcessor
from dlk.utils.register import register, register_module_name

logger = logging.getLogger(__name__)


@st.cache_resource
def get_process(config: str) -> PreProcessor:
    """

    Args:
        config: the path to process config

    Returns:
        process instance

    """
    return PreProcessor(config, stage="online")


@st.cache_resource
def get_predict(config: str, checkpoint: str) -> OnlinePredict:
    """

    Args:
        config: the path to predict config
        checkpoint: the path to checkpoint

    Returns:
        predict instance

    """
    return OnlinePredict(config, checkpoint)


@st.cache_resource
def prepare_display(dump_display_config: str) -> Display:
    """load the display instance from the config

    Args:
        dump_display_config: the json dumps display config

    Returns:
        Display
    """

    configs = Parser(json.loads(dump_display_config)).parser_init()
    assert len(configs) == 1, f"You should not use '_search' for demo"
    display_config = configs[0]["@display"]
    display_name = register_module_name(display_config._module_name)
    display = register.get("display", display_name)(display_config)

    return display


if "predict" not in st.session_state:
    st.session_state.predict = None

if "process" not in st.session_state:
    st.session_state.process = None


class Demo(object):
    """Demo"""

    def __init__(
        self,
        display_config: Union[str, Dict],
        process_config: str = "/path/to/config",
        fit_config: str = "/path/to/config",
        checkpoint: str = "/path/to/checkpoint",
        scrolling=True,
    ):
        super(Demo, self).__init__()
        if isinstance(display_config, str):
            with open(display_config, "r") as f:
                display_config = hjson.load(f, object_pairs_hook=dict)
        self.display = prepare_display(json.dumps(display_config, sort_keys=True))
        self.scrolling = scrolling

        with st.sidebar:
            st.title("DLK Demo")
            process_config_value = st.text_input(
                "Process Config",
                value=None if process_config == "/path/to/config" else process_config,
                placeholder=process_config,
            )

            fit_config_value = st.text_input(
                "Main Config",
                value=None if fit_config == "/path/to/config" else fit_config,
                placeholder=fit_config,
            )
            checkpoint_path_value = st.text_input(
                "Checkpoint Path",
                value=None if checkpoint == "/path/to/checkpoint" else checkpoint,
                placeholder=checkpoint,
            )

            if st.button("Load Model"):
                self.load_model(
                    process_config_value, fit_config_value, checkpoint_path_value
                )

            if st.button("Clear Cache"):
                st.cache_resource.clear()

        st.header(self.display.config.title)
        if self.display.config.help:
            st.caption(self.display.config.help)
        st.markdown("--------")
        input_names = asdict(self.display.config.input)
        inputs = {}
        inputs["uuid"] = str(uuid4())
        inputs.update(asdict(self.display.config.hold))
        for input in input_names:
            if input_names[input] == "text":
                st.subheader(f"{input}:")
                inputs[input] = st.text_area(
                    label=" ",
                    placeholder=f"Please input the {input}",
                    value="",
                    height=3,
                )
            elif input_names[input] == "image":
                st.subheader(f"{input}:")
                image = st.file_uploader(label=" ", type=["png", "jpg", "jpeg"])
                if image:
                    image = Image.open(io.BytesIO(image.read()))
                    st.image(image, caption=f"Uploaded {input}", use_column_width=True)
                    inputs[input] = image
                else:
                    st.text("Please upload the right image")
            else:
                st.text("Please check the input type")
        if st.button("Submit"):
            self.generate_response(pd.DataFrame([inputs]))

    def load_model(self, process_config_value, fit_config_value, checkpoint_path_value):
        st.session_state.process = get_process(process_config_value)
        logger.info("Load Processor Done")
        st.session_state.predict = get_predict(fit_config_value, checkpoint_path_value)
        logger.info("Load Predictor Done")

    def generate_response(self, input_df):
        if (
            st.session_state.predict is not None
            and st.session_state.process is not None
        ):
            st.markdown("--------")
            st.subheader(f"{self.display.config.output_head}:")
            processed_data = st.session_state.process.fit(input_df)
            result = st.session_state.predict.predict(processed_data)
            if self.display.config.render_type == "html":
                components.html(
                    self.display.display(result[0]),
                    width=int(self.display.config.render_width * 1.1),
                    height=int(self.display.config.render_height * 1.1),
                    scrolling=self.scrolling,
                )
            elif self.display.config.render_type == "text":
                st.text_area(
                    "", self.display.display(result[0]), label_visibility="hidden"
                )
            elif self.display.config.render_type == "image":
                st.image(self.display.display(result[0]))
            else:
                st.text("Please check the render type")
            with st.expander("Json Result"):
                st.text(json.dumps(result[0], indent=4))
        else:
            st.text("Please load the model first")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import abc
import importlib
import os

from lightning import LightningDataModule

from dlk.utils.import_module import import_module_dir


class IBaseDataModule(LightningDataModule):
    """docstring for IBaseDataModule"""

    def __init__(self):
        super(IBaseDataModule, self).__init__()

    def train_dataloader(self):
        """

        Raises:
            NotImplementedError

        """
        raise NotImplementedError(
            f"You must implementation the train_dataloader for your own datamodule."
        )

    def predict_dataloader(self):
        """

        Raises:
            NotImplementedError

        """
        raise NotImplementedError(
            f"You must implementation the predict_dataloader for your own datamodule."
        )

    def val_dataloader(self):
        """

        Raises:
            NotImplementedError

        """
        raise NotImplementedError(
            f"You must implementation the val_dataloader for your own datamodule."
        )

    def test_dataloader(self):
        """

        Raises:
            NotImplementedError

        """
        raise NotImplementedError(
            f"You must implementation the test_dataloader for your own datamodule."
        )

    @abc.abstractmethod
    def online_dataloader(self):
        """

        Raises:
            NotImplementedError

        """
        raise NotImplementedError(
            f"You must implementation the online_dataloader for your own datamodule."
        )


# automatically import any Python files in the models directory
datamodule_dir = os.path.dirname(__file__)
import_module_dir(datamodule_dir, "dlk.data.datamodule")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
    dataclass,
)

from dlk.utils.import_module import import_module_dir


@dataclass
class DisplayConfigBase(Base):
    """the display config"""

    class Input:
        pass

    class Hold:
        pass

    title = StrField("", help="the title of the demo")
    help = StrField("", help="the help infomation of the demo")
    input = NestField(value=Input, help="the input config")
    hold = NestField(value=Hold, help="the hold value")
    output_head = StrField(value="Output", help="the output head")
    render_height = IntField(value=500, help="the display height")
    render_width = IntField(value=900, help="the display width")
    render_type = StrField(
        value="html",
        options=["html", "image", "text"],
        help="the render type of the display",
    )


class Display(object):
    """Display Base"""

    def __init__(self, config: DisplayConfigBase):
        super(Display, self).__init__()
        self.config = config

    def display(self, result: Dict) -> str:
        """

        Args:
            result (Dict): TODO

        Returns: TODO

        """
        raise NotImplementedError

    def __call__(self, result: Dict) -> str:
        """display the result

        Args:
            result:

        Returns:
            html content

        """
        assert isinstance(result, dict), "result should be a dict"

        return self.display(result)


imodel_dir = os.path.dirname(__file__)
import_module_dir(imodel_dir, "dlk.display")
